"""
This module provides methods to calculate NLL(Negative Log-Likelihood) as well as its derivatives.
"""

import math
import warnings
from itertools import repeat as _loop_generator

import numpy as np

from ..config import get_config
from ..data import data_merge, data_shape, data_split, split_generator
from ..tensorflow_wrapper import tf
from ..utils import time_print
from ..variable import Variable


def get_shape(x):
    if hasattr(x, "shape"):
        return x.shape
    return ()


def _resolution_shape(x):
    shape = get_shape(x)
    if shape:
        return shape[-1]
    return 1


def _batch_sum(f, data_i, weight_i, trans, resolution_size, args, kwargs):
    weight_shape = (-1, min(_resolution_shape(weight_i), resolution_size))
    part_y = f(data_i, *args, **kwargs)
    weight_i = tf.cast(weight_i, part_y.dtype)
    rw = tf.reshape(weight_i, weight_shape)
    part_y = weight_i * part_y
    part_y = tf.reshape(part_y, (-1, resolution_size))
    part_y = tf.reduce_sum(part_y, axis=-1)
    event_w = tf.reduce_sum(rw, axis=-1)
    part_y = part_y / event_w
    part_y = trans(part_y)
    y_i = tf.reduce_sum(event_w * part_y)
    return y_i


def sum_gradient(
    f,
    data,
    var,
    weight=1.0,
    trans=tf.identity,
    resolution_size=1,
    args=(),
    kwargs=None,
):
    """
    NLL is the sum of trans(f(data)):math:`*`weight; gradient is the derivatives for each variable in ``var``.

    :param f: Function. The amplitude PDF.
    :param data: Data array
    :param var: List of strings. Names of the trainable variables in the PDF.
    :param weight: Weight factor for each data point. It's either a real number or an array of the same shape with ``data``.
    :param trans: Function. Transformation of ``data`` before multiplied by ``weight``.
    :param kwargs: Further arguments for ``f``.
    :return: Real number NLL, list gradient
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    ys = []
    gs = []
    for data_i, weight_i in zip(data, weight):
        with tf.GradientTape() as tape:
            y_i = _batch_sum(
                f, data_i, weight_i, trans, resolution_size, args, kwargs
            )
        g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
        ys.append(y_i)
        gs.append(g_i)
    nll = sum(ys)
    # print("ll0:,", nll)
    g = list(map(sum, zip(*gs)))
    return nll, g


def sum_hessian(
    f,
    data,
    var,
    weight=1.0,
    trans=tf.identity,
    resolution_size=1,
    args=(),
    kwargs=None,
):
    """
    The parameters are the same with ``sum_gradient()``, but this function will return hessian as well,
    which is the matrix of the second-order derivative.

    :return: Real number NLL, list gradient, 2-D list hessian
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    y_s = []
    g_s = []
    h_s = []
    for data_i, weight_i in zip(data, weight):
        with tf.GradientTape(persistent=True) as tape0:
            with tf.GradientTape() as tape:
                y_i = _batch_sum(
                    f, data_i, weight_i, trans, resolution_size, args, kwargs
                )
            g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
        h_s_i = []
        # h_s_i  = tape0.jacobian(tf.stack(g_i), var, unconnected_gradients="zero")
        for gi in g_i:
            # 2nd order derivative
            h_s_i.append(tape0.gradient(gi, var, unconnected_gradients="zero"))
        del tape0
        y_s.append(y_i)
        g_s.append(g_i)
        h_s.append(h_s_i)
    nll = tf.reduce_sum(y_s)
    # print("ll: ", nll)
    g = tf.reduce_sum(g_s, axis=0)
    h = tf.reduce_sum(h_s, axis=0)
    # h = [[sum(j) for j in zip(*i)] for i in h_s]
    return nll, g, h


def sum_grad_hessp(
    f,
    p,
    data,
    var,
    weight=1.0,
    trans=tf.identity,
    resolution_size=1,
    args=(),
    kwargs=None,
):
    """
    The parameters are the same with ``sum_gradient()``, but this function will return hessian as well,
    which is the matrix of the second-order derivative.

    :return: Real number NLL, list gradient, 2-D list hessian
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    y_s = []
    g_s = []
    h_s = []
    from tensorflow.python.eager import forwardprop

    for data_i, weight_i in zip(data, weight):
        with forwardprop.ForwardAccumulator(var, list(p)) as acc:
            with tf.GradientTape() as tape:
                y_i = _batch_sum(
                    f, data_i, weight_i, trans, resolution_size, args, kwargs
                )
            g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
        hessp = acc.jvp(g_i, unconnected_gradients="zero")
        y_s.append(y_i)
        g_s.append(g_i)
        h_s.append(hessp)
        # print(hessp)
    nll = tf.reduce_sum(y_s)
    # print("ll: ", nll)
    g = tf.reduce_sum(g_s, axis=0)
    h = tf.reduce_sum(h_s, axis=0)
    # print(h)
    # h = [[sum(j) for j in zip(*i)] for i in h_s]
    return nll, g, h


def sum_gradient_new(
    amp,
    data,
    mcdata,
    weight,
    mcweight,
    var,
    trans=tf.math.log,
    w_flatmc=lambda: 0,
    args=(),
    kwargs=None,
):
    """
    NLL is the sum of trans(f(data)):math:`*`weight; gradient is the derivatives for each variable in ``var``.

    :param f: Function. The amplitude PDF.
    :param data: Data array
    :param var: List of strings. Names of the trainable variables in the PDF.
    :param weight: Weight factor for each data point. It's either a real number or an array of the same shape with ``data``.
    :param trans: Function. Transformation of ``data`` before multiplied by ``weight``.
    :param kwargs: Further arguments for ``f``.
    :return: Real number NLL, list gradient
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    ys = []
    # gs = []
    ymc = []
    with tf.GradientTape() as tape:
        for mcdata_i, mcweight_i in zip(mcdata, mcweight):
            part_y = amp(mcdata_i, *args, **kwargs)
            y_i = tf.reduce_sum(tf.cast(mcweight_i, part_y.dtype) * part_y)
            ymc.append(y_i)
        int_dt = tf.reduce_sum(ymc)
        for data_i, weight_i in zip(data, weight):
            wmc = w_flatmc()
            part_y = (amp(data_i, *args, **kwargs) / int_dt + wmc) / (1 + wmc)
            part_y = trans(part_y)
            y_i = tf.reduce_sum(tf.cast(weight_i, part_y.dtype) * part_y)
            ys.append(y_i)
        nll = -tf.reduce_sum(ys)
    g = tape.gradient(nll, var, unconnected_gradients="zero")
    return nll, g


def sum_hessian_new(
    amp,
    data,
    mcdata,
    weight,
    mcweight,
    var,
    trans=tf.math.log,
    w_flatmc=lambda: 0,
    args=(),
    kwargs=None,
):
    """
    The parameters are the same with ``sum_gradient()``, but this function will return hessian as well,
    which is the matrix of the second-order derivative.

    :return: Real number NLL, list gradient, 2-D list hessian
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    ys = []
    ymc = []
    with tf.GradientTape(persistent=True) as tape0:
        with tf.GradientTape() as tape:
            for mcdata_i, mcweight_i in zip(mcdata, mcweight):
                part_y = amp(mcdata_i, *args, **kwargs)
                y_i = tf.reduce_sum(tf.cast(mcweight_i, part_y.dtype) * part_y)
                ymc.append(y_i)
            int_dt = tf.reduce_sum(ymc)
            for data_i, weight_i in zip(data, weight):
                wmc = w_flatmc()
                part_y = (amp(data_i, *args, **kwargs) / int_dt + wmc) / (
                    1 + wmc
                )
                part_y = trans(part_y)
                y_i = tf.reduce_sum(tf.cast(weight_i, part_y.dtype) * part_y)
                ys.append(y_i)
            nll = -tf.reduce_sum(ys)
        gradient = tape.gradient(nll, var, unconnected_gradients="zero")
    hessian = []
    for gi in gradient:
        # 2nd order derivative
        hessian.append(tape0.gradient(gi, var, unconnected_gradients="zero"))
    del tape0
    return nll, gradient, hessian


def clip_log(x, _epsilon=1e-6):
    """clip log to allowed large value"""
    x_cut = tf.where(x > _epsilon, x, tf.ones_like(x) * _epsilon)
    b_t = tf.math.log(x_cut)
    delta_x = x - _epsilon
    b_f = (
        np.log(_epsilon) + delta_x / _epsilon - (delta_x / _epsilon) ** 2 / 2.0
    )
    # print("$$$", tf.where(x < _epsilon).numpy().tolist())
    return tf.where(x > _epsilon, b_t, b_f)


class BaseModel(object):
    """
    This class implements methods to calculate NLL as well as its derivatives for an amplitude model. It may include
    data for both signal and background.

    :param signal: Signal Model
    """

    def __init__(self, signal, resolution_size=1):
        self.signal = signal
        self.Amp = signal
        self.vm = signal.vm
        self.resolution_size = resolution_size

    def nll(self, data, mcdata):
        """Negative log-Likelihood"""
        weight = data.get("weight", tf.ones((data_shape(data),)))
        sw = tf.reduce_sum(weight)
        rw = tf.reshape(weight, (-1, self.resolution_size))
        amp_s2 = self.signal(data) * weight
        amp_s2 = tf.reshape(amp_s2, (-1, self.resolution_size))
        amp_s2 = tf.reduce_sum(amp_s2, axis=-1)
        weight = tf.reduce_sum(rw, axis=-1)
        ln_data = clip_log(amp_s2 / weight)
        mc_weight = mcdata.get("weight", tf.ones((data_shape(mcdata),)))
        int_mc = tf.reduce_sum(
            mc_weight * self.signal(mcdata)
        ) / tf.reduce_sum(mc_weight)
        alpha = sw / tf.reduce_sum(weight ** 2)
        return -alpha * (
            tf.reduce_sum(weight * ln_data) - sw * tf.math.log(int_mc)
        )

    def sum_nll_grad_bacth(self, data):
        weight = [i.get("weight", tf.ones((data_shape(i),))) for i in data]
        ln_data, g_ln_data = sum_gradient(
            self.signal,
            data,
            self.signal.trainable_variables,
            weight=weight,
            trans=clip_log,
            resolution_size=self.resolution_size,
        )
        return -ln_data, [-i for i in g_ln_data]

    def sum_log_integral_grad_batch(self, mcdata, ndata):
        mc_weight = [i["weight"] for i in mcdata]
        int_mc, g_int_mc = sum_gradient(
            self.signal,
            mcdata,
            self.signal.trainable_variables,
            weight=mc_weight,
        )
        return tf.math.log(int_mc) * ndata, [
            ndata / int_mc * i for i in g_int_mc
        ]

    def nll_grad(self, data, mcdata, batch=65000):
        weight = data.get("weight", tf.ones((data_shape(data),)))
        weight_rw = tf.reduce_sum(
            tf.reshape(weight, (-1, self.resolution_size)), axis=-1
        )
        alpha = tf.reduce_sum(weight_rw) / tf.reduce_sum(weight_rw ** 2)
        weight = alpha * weight
        assert (
            batch % self.resolution_size == 0
        ), "batch size should be the multiple of resolution_size"
        ln_data, g_ln_data = sum_gradient(
            self.signal,
            split_generator(data, batch),
            self.signal.trainable_variables,
            weight=split_generator(weight, batch),
            trans=clip_log,
            resolution_size=self.resolution_size,
        )
        mc_weight = mcdata.get("weight", tf.ones((data_shape(mcdata),)))
        mc_weight = mc_weight / tf.reduce_sum(mc_weight)
        int_mc, g_int_mc = sum_gradient(
            self.signal,
            split_generator(mcdata, batch),
            self.signal.trainable_variables,
            weight=data_split(mc_weight, batch),
        )

        sw = tf.cast(tf.reduce_sum(weight), ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )
        nll = -ln_data + sw * tf.math.log(int_mc)
        return nll, g

    @property
    def trainable_variables(self):
        return self.signal.trainable_variables

    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        weight = list(weight)
        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
        ln_data, g_ln_data = sum_gradient(
            self.signal,
            data,
            self.signal.trainable_variables,
            weight=weight,
            trans=clip_log,
            resolution_size=self.resolution_size,
        )
        int_mc, g_int_mc = sum_gradient(
            self.signal,
            mcdata,
            self.signal.trainable_variables,
            weight=mc_weight,
        )

        sw = tf.cast(sw, ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )
        nll = -ln_data + sw * tf.math.log(int_mc)
        return nll, g

    def grad_hessp_batch(self, p, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        if not hasattr(self, "hess_product_vector_i"):
            self.hess_product_vector_i = [tf.Variable(i) for i in p]
        for i, j in zip(self.hess_product_vector_i, p):
            i.assign(j)
        weight = list(weight)
        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
        ln_data, g_ln_data, hessp_ln_data = sum_grad_hessp(
            self.signal,
            self.hess_product_vector_i,
            data,
            self.signal.trainable_variables,
            weight=weight,
            trans=clip_log,
            resolution_size=self.resolution_size,
        )

        # print("hessp_ln_data",hessp_ln_data)

        int_mc, g_int_mc, hessp_int_mc = sum_grad_hessp(
            self.signal,
            self.hess_product_vector_i,
            mcdata,
            self.signal.trainable_variables,
            weight=mc_weight,
        )

        # print("hessp_int_mc", hessp_int_mc)

        sw = tf.cast(sw, ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )

        g_int_mc = np.array(g_int_mc)
        hessp2 = sw * (
            hessp_int_mc / int_mc
            - g_int_mc * np.dot(p, g_int_mc) / int_mc ** 2
        )
        # print("hessp2", hessp2)
        # print("ret", g, hessp2 - hessp_ln_data)
        return g, hessp2 - hessp_ln_data

    def nll_grad_hessian(self, data, mcdata, batch=25000):
        """
        The parameters are the same with ``self.nll()``, but it will return Hessian as well.

        :return NLL: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return Hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        assert (
            batch % self.resolution_size == 0
        ), "batch size should be the multiple of resolution_size"
        weight = data.get("weight", tf.ones((data_shape(data),)))
        mc_weight = mcdata.get("weight", tf.ones((data_shape(mcdata),)))
        mc_weight = mc_weight / tf.reduce_sum(mc_weight)
        weight_rw = tf.reduce_sum(
            tf.reshape(weight, (-1, self.resolution_size)), axis=-1
        )
        alpha = tf.reduce_sum(weight_rw) / tf.reduce_sum(weight_rw ** 2)
        weight = alpha * weight
        sw = tf.reduce_sum(weight)
        ln_data, g_ln_data, h_ln_data = sum_hessian(
            self.signal,
            split_generator(data, batch),
            self.signal.trainable_variables,
            weight=split_generator(weight, batch),
            trans=clip_log,
            resolution_size=self.resolution_size,
        )
        int_mc, g_int_mc, h_int_mc = sum_hessian(
            self.signal,
            split_generator(mcdata, batch),
            self.signal.trainable_variables,
            weight=split_generator(mc_weight, batch),
        )

        n_var = len(g_ln_data)
        nll = -ln_data + sw * tf.math.log(int_mc)
        g = -g_ln_data + sw * g_int_mc / int_mc

        g_int_mc = g_int_mc / int_mc
        g_outer = tf.reshape(g_int_mc, (-1, 1)) * tf.reshape(g_int_mc, (1, -1))

        h = -h_ln_data - sw * g_outer + sw / int_mc * h_int_mc
        # print("nll: ", nll)
        return nll, g, h

    def set_params(self, var):
        """
        It has interface to ``Amplitude.set_params()``.
        """
        self.Amp.set_params(var)

    def get_params(self, trainable_only=False):
        """
        It has interface to ``Amplitude.get_params()``.
        """
        return self.Amp.get_params(trainable_only)


class Model(object):
    """
    This class implements methods to calculate NLL as well as its derivatives for an amplitude model. It may include
    data for both signal and background.

    :param amp: ``AllAmplitude`` object. The amplitude model.
    :param w_bkg: Real number. The weight of background.
    """

    def __init__(self, amp, w_bkg=1.0, resolution_size=1):
        self.model = BaseModel(amp, resolution_size=resolution_size)
        self.Amp = amp
        self.w_bkg = w_bkg
        self.vm = amp.vm
        self.resolution_size = self.model.resolution_size

    def get_weight_data(self, data, weight=None, bg=None, alpha=True):
        """
        Blend data and background data together multiplied by their weights.

        :param data: Data array
        :param weight: Weight for data
        :param bg: Data array for background
        :param alpha: Boolean. If it's true, ``weight`` will be multiplied by a factor :math:`\\alpha=`???
        :return: Data, weight. Their length both equals ``len(data)+len(bg)``.
        """
        if weight is None:
            weight = data.get("weight", 1.0)
        if isinstance(weight, float):
            n_data = data_shape(data)
            weight = tf.convert_to_tensor(
                [weight] * n_data, dtype=get_config("dtype")
            )
        if bg is not None:
            n_bg = data_shape(bg)
            data = data_merge(data, bg)
            bg_weight = bg.get("weight", None)
            if bg_weight is None:
                bg_weight = tf.convert_to_tensor(
                    [-self.w_bkg] * n_bg, dtype=get_config("dtype")
                )
            else:
                bg_weight = tf.convert_to_tensor(
                    bg_weight, dtype=get_config("dtype")
                )
            weight = tf.concat([weight, bg_weight], axis=0)
        # print(weight.shape)
        if alpha:
            weight_r = tf.reshape(weight, (-1, self.resolution_size))
            weight_r = tf.reduce_sum(weight_r, axis=-1)
            alpha = tf.reduce_sum(weight_r) / tf.reduce_sum(
                weight_r * weight_r
            )
            return data, alpha * weight
        return data, weight

    def mix_data_bakcground(self, data, bg):
        ret, weight = self.get_weight_data(
            data, weight=data.get_weight(), bg=bg, alpha=True
        )
        ret["weight"] = weight
        return ret

    def sum_nll_grad_bacth(self, data):
        return self.model.sum_nll_grad_bacth(data)

    def sum_log_integral_grad_batch(self, mcdata, ndata):
        return self.model.sum_log_integral_grad_batch(mcdata, ndata)

    def nll(
        self,
        data,
        mcdata,
        weight: tf.Tensor = 1.0,
        batch=None,
        bg=None,
        mc_weight=1.0,
    ):
        """
        Calculate NLL.

        .. math::
          -\\ln L = -\\sum_{x_i \\in data } w_i \\ln f(x_i;\\theta_k) +  (\\sum w_j ) \\ln \\sum_{x_i \\in mc } f(x_i;\\theta_k)

        :param data: Data array
        :param mcdata: MCdata array
        :param weight: Weight of data???
        :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
        :param bg: Background data array. It can be set to ``None`` if there is no such thing.
        :return: Real number. The value of NLL.
        """
        data, weight = self.get_weight_data(data, weight, bg=bg)
        if isinstance(mc_weight, float):
            mc_weight = tf.convert_to_tensor(
                [mc_weight] * data_shape(mcdata), dtype="float64"
            )
        return self.model.nll(
            {**data, "weight": weight}, {**mcdata, "weight": mc_weight}
        )

    def nll_grad(
        self, data, mcdata, weight=1.0, batch=65000, bg=None, mc_weight=1.0
    ):
        """
        Calculate NLL and its gradients.

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        The parameters are the same with ``self.nll()``, but it will return gradients as well.

        :return NLL: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        data, weight = self.get_weight_data(data, weight, bg=bg)
        if isinstance(mc_weight, float):
            mc_weight = tf.convert_to_tensor(
                [mc_weight] * data_shape(mcdata), dtype="float64"
            )
        return self.model.nll_grad(
            {**data, "weight": weight},
            {**mcdata, "weight": mc_weight},
            batch=batch,
        )

    # @tf.function
    def grad_hessp_batch(self, p, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        data_i = ({**i, "weight": j} for i, j in zip(data, weight))
        mcdata_i = ({**i, "weight": j} for i, j in zip(mcdata, mc_weight))
        return self.model.grad_hessp_batch(
            p, data_i, mcdata_i, weight, mc_weight
        )

    # @tf.function
    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        data_i = ({**i, "weight": j} for i, j in zip(data, weight))
        mcdata_i = ({**i, "weight": j} for i, j in zip(mcdata, mc_weight))
        return self.model.nll_grad_batch(data_i, mcdata_i, weight, mc_weight)

    def nll_grad_hessian(
        self, data, mcdata, weight=1.0, batch=24000, bg=None, mc_weight=1.0
    ):
        """
        The parameters are the same with ``self.nll()``, but it will return Hessian as well.

        :return NLL: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return Hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        data, weight = self.get_weight_data(data, weight, bg=bg)
        if isinstance(mc_weight, float):
            mc_weight = tf.convert_to_tensor(
                [mc_weight] * data_shape(mcdata), dtype="float64"
            )
        data_i = {**data, "weight": weight}
        mcdata_i = {**mcdata, "weight": mc_weight}
        return self.model.nll_grad_hessian(data_i, mcdata_i, batch=batch)

    def set_params(self, var):
        """
        It has interface to ``Amplitude.set_params()``.
        """
        self.Amp.set_params(var)

    def get_params(self, trainable_only=False):
        """
        It has interface to ``Amplitude.get_params()``.
        """
        return self.Amp.get_params(trainable_only)


class Model_new(Model):
    """
    This class implements methods to calculate NLL as well as its derivatives for an amplitude model. It may include
    data for both signal and background.

    :param amp: ``AllAmplitude`` object. The amplitude model.
    :param w_bkg: Real number. The weight of background.
    """

    def __init__(self, amp, w_bkg=1.0, w_inmc=0, float_wmc=False):
        super(Model_new, self).__init__(amp, w_bkg)
        # self.w_inmc = w_inmc
        self.w_inmc = Variable(
            "weight_injectMC", value=w_inmc, vm=self.Amp.vm, overwrite=False
        )
        if not float_wmc:
            self.w_inmc.fixed()

    def get_weight_data(
        self, data, weight=1.0, bg=None, inmc=None, alpha=True
    ):
        """
        Blend data and background data together multiplied by their weights.

        :param data: Data array
        :param weight: Weight for data
        :param bg: Data array for background
        :param alpha: Boolean. If it's true, ``weight`` will be multiplied by a factor :math:`\\alpha=`???
        :return: Data, weight. Their length both equals ``len(data)+len(bg)``.
        """
        n_data = data_shape(data)
        if isinstance(weight, float):
            weight = tf.convert_to_tensor(
                [weight] * n_data, dtype=get_config("dtype")
            )
        if bg is not None:
            n_bg = data_shape(bg)
            data = data_merge(data, bg)
            bg_weight = tf.convert_to_tensor(
                [-self.w_bkg] * n_bg, dtype=get_config("dtype")
            )
            weight = tf.concat([weight, bg_weight], axis=0)
        if inmc is not None:
            n_inmc = data_shape(inmc)
            data = data_merge(data, inmc)
            wmc = self.w_inmc() * n_data / n_inmc
            inmc_weight = tf.convert_to_tensor(
                [wmc] * n_inmc, dtype=get_config("dtype")
            )
            weight = tf.concat([weight, inmc_weight], axis=0)

        if alpha:
            alpha = tf.reduce_sum(weight) / tf.reduce_sum(
                weight * weight
            )  # correct with inject MC?
            return data, alpha * weight
        return data, weight

    def nll(self, data, mcdata, weight: tf.Tensor = 1.0, batch=None, bg=None):
        """
        Calculate NLL.

        .. math::
          -\\ln L = -\\sum_{x_i \\in data } w_i \\ln f(x_i;\\theta_k) +  (\\sum w_j ) \\ln \\sum_{x_i \\in mc } f(x_i;\\theta_k)

        :param data: Data array
        :param mcdata: MCdata array
        :param weight: Weight of data???
        :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
        :param bg: Background data array. It can be set to ``None`` if there is no such thing.
        :return: Real number. The value of NLL.
        """
        data, weight = self.get_weight_data(data, weight, bg=bg)
        sw = tf.reduce_sum(weight)
        ln_data = tf.math.log(self.Amp(data))
        int_mc = tf.math.log(tf.reduce_mean(self.Amp(mcdata)))
        nll_0 = -tf.reduce_sum(tf.cast(weight, ln_data.dtype) * ln_data)
        return nll_0 + tf.cast(sw, int_mc.dtype) * int_mc

    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad_new``
        """
        # N_data = 2525
        # N_flatmc = 3106
        # w_flatmc = 0#878/2525#2889/2525#3106/2525
        nll, g = sum_gradient_new(
            self.Amp,
            data,
            mcdata,
            weight,
            mc_weight,
            self.Amp.trainable_variables,
            clip_log,
            self.w_inmc,
        )
        # print("@@@",nll,np.array(g).tolist())
        return nll, g

    def nll_grad_hessian(self, data, mcdata, weight, mc_weight):
        """
        The parameters are the same with ``self.nll()``, but it will return Hessian as well.

        :return NLL: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return Hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        nll, g, h = sum_hessian_new(
            self.Amp,
            data,
            mcdata,
            weight,
            mc_weight,
            self.Amp.trainable_variables,
            clip_log,
            self.w_inmc,
        )
        h = tf.stack(h)
        return nll, g, h


class GaussianConstr(object):
    def __init__(self, vm, constraint={}):
        self.vm = vm
        self.constraint = {}
        self.update(constraint)

    def update(self, constraint={}):
        for i in constraint:
            if not i in self.vm.trainable_vars:
                warnings.warn(
                    "Constraint {} is useless to fitting because it's not trainable".format(
                        i
                    )
                )
        self.constraint.update(constraint)

    def get_constrain_term(self):
        r"""
        constraint: Gauss(mean,sigma)
          by add a term :math:`\frac{(\theta_i-\bar{\theta_i})^2}{2\sigma^2}`
        """
        term = 0.0
        for i in self.constraint:
            pi = self.constraint[i]
            assert isinstance(pi, tuple) or isinstance(pi, list)
            assert len(pi) == 2
            mean, sigma = pi
            var = self.vm.variables[i]
            term += (var - mean) ** 2 / (sigma ** 2) / 2
        return term

    def get_constrain_grad(self):
        r"""
        constraint: Gauss(mean,sigma)
          by add a term :math:`\frac{d}{d\theta_i}\frac{(\theta_i-\bar{\theta_i})^2}{2\sigma^2} = \frac{\theta_i-\bar{\theta_i}}{\sigma^2}`
        """
        g_dict = {}
        for i in self.constraint:
            if not i in self.vm.trainable_vars:
                continue
            pi = self.constraint[i]
            assert isinstance(pi, tuple) or isinstance(pi, list)
            assert len(pi) == 2
            mean, sigma = pi
            var = self.vm.variables[i]
            g_dict[i] = (var - mean) / (sigma ** 2)  # 1st differentiation
        grad = []
        for i in self.vm.trainable_vars:
            if i in g_dict:
                grad.append(g_dict[i])
            else:
                grad.append(0.0)
        return np.array(grad)

    def get_constrain_hessian(self):
        """the constrained parameter's 2nd differentiation"""
        h_dict = {}
        for i in self.constraint:
            if not i in self.vm.trainable_vars:
                continue
            pi = self.constraint[i]
            assert isinstance(pi, tuple) or isinstance(pi, list)
            assert len(pi) == 2
            mean, sigma = pi
            var = self.vm.variables[i]
            h_dict[i] = 1 / (sigma ** 2)  # 2nd differentiation
        nv = len(self.vm.trainable_vars)
        hessian = np.zeros([nv, nv])
        for v, i in zip(self.vm.trainable_vars, range(nv)):
            if v in h_dict:
                hessian[i, i] = h_dict[v]
        return np.array(hessian)


class ConstrainModel(Model):
    """
    negative log likelihood model with constrains

    """

    def __init__(self, amp, w_bkg=1.0, constrain={}):
        super(ConstrainModel, self).__init__(amp, w_bkg)
        self.constrain = (
            constrain  # priori gauss constrain for the fitting parameters
        )

    def get_constrain_term(self):  # the priori constrain term added to NLL
        r"""
        constrain: Gauss(mean,sigma)
          by add a term :math:`\frac{(\theta_i-\bar{\theta_i})^2}{2\sigma^2}`

        """
        t_var = self.Amp.trainable_variables
        t_var_name = [i.name for i in t_var]
        var_dict = dict(zip(t_var_name, t_var))
        nll = 0.0
        for i in self.constrain:
            if not i in var_dict:
                break
            pi = self.constrain[i]
            if isinstance(pi, tuple) and len(pi) == 2:
                mean, sigma = pi
                var = var_dict[i]
                nll += (var - mean) ** 2 / (sigma ** 2) / 2
        return nll

    def get_constrain_grad(
        self,
    ):  # the constrained parameter's 1st differentiation
        r"""
        constrain: Gauss(mean,sigma)
          by add a term :math:`\frac{d}{d\theta_i}\frac{(\theta_i-\bar{\theta_i})^2}{2\sigma^2} = \frac{\theta_i-\bar{\theta_i}}{\sigma^2}`

        """
        t_var = self.Amp.trainable_variables
        t_var_name = [i.name for i in t_var]
        var_dict = dict(zip(t_var_name, t_var))
        g_dict = {}
        for i in self.constrain:
            if not i in var_dict:
                break
            pi = self.constrain[i]
            if isinstance(pi, tuple) and len(pi) == 2:
                mean, sigma = pi
                var = var_dict[i]
                g_dict[i] = (var - mean) / (sigma ** 2)  # 1st differentiation
        nll_g = []
        for i in t_var_name:
            if i in g_dict:
                nll_g.append(g_dict[i])
            else:
                nll_g.append(0.0)
        return nll_g

    def get_constrain_hessian(self):
        """the constrained parameter's 2nd differentiation"""
        t_var = self.Amp.trainable_variables
        t_var_name = [i.name for i in t_var]
        var_dict = dict(zip(t_var_name, t_var))
        g_dict = {}
        for i in self.constrain:
            if not i in var_dict:
                break
            pi = self.constrain[i]
            if isinstance(pi, tuple) and len(pi) == 2:
                mean, sigma = pi
                var = var_dict[i]
                g_dict[i] = 1 / (sigma ** 2)  # 2nd differentiation
        nll_g = []
        for i in t_var_name:
            if i in g_dict:
                nll_g.append(g_dict[i])
            else:
                nll_g.append(0.0)
        return np.diag(nll_g)

    def nll(self, data, mcdata, weight=1.0, bg=None, batch=None):
        r"""
        calculate negative log-likelihood

        .. math::
          -\ln L = -\sum_{x_i \in data } w_i \ln f(x_i;\theta_i) +  (\sum w_i ) \ln \sum_{x_i \in mc } f(x_i;\theta_i) + cons

        """
        nll_0 = super(ConstrainModel, self).nll(
            data, mcdata, weight=weight, batch=batch, bg=bg
        )
        cons = self.get_constrain_term()
        return nll_0 + cons

    def nll_gradient(self, data, mcdata, weight=1.0, batch=None, bg=None):
        r"""
        calculate negative log-likelihood with gradient

        .. math::
          \frac{\partial }{\partial \theta_i }(-\ln L) = -\sum_{x_i \in data } w_i \frac{\partial }{\partial \theta_i } \ln f(x_i;\theta_i) +
          \frac{\sum w_i }{\sum_{x_i \in mc }f(x_i;\theta_i)} \sum_{x_i \in mc } \frac{\partial }{\partial \theta_i } f(x_i;\theta_i) + cons

        """
        cons_grad = self.get_constrain_grad()  # the constrain term
        cons = self.get_constrain_term()
        nll0, g0 = super(ConstrainModel, self).nll_grad(
            data, mcdata, weight=weight, batch=batch, bg=bg
        )
        nll = nll0 + cons
        g = [cons_grad[i] + g0[i] for i in range(len(g0))]
        return nll, g


class FCN(object):
    """
    This class implements methods to calculate the NLL as well as its derivatives for a general function.

    :param model: Model object.
    :param data: Data array.
    :param mcdata: MCdata array.
    :param bg: Background array.
    :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
    """

    def __init__(
        self,
        model,
        data,
        mcdata,
        bg=None,
        batch=65000,
        inmc=None,
        gauss_constr={},
    ):
        self.model = model
        self.vm = model.vm
        self.n_call = 0
        self.n_grad = 0
        self.cached_nll = None
        if inmc is None:
            data, weight = self.model.get_weight_data(data, bg=bg)
            print("Using Model_old")
        else:
            data, weight = self.model.get_weight_data(data, bg=bg, inmc=inmc)
            print("Using Model_new")
        n_mcdata = data_shape(mcdata)
        self.alpha = tf.reduce_sum(weight) / tf.reduce_sum(weight * weight)
        self.weight = weight
        self.data = data
        self.batch_data = list(split_generator(data, batch))
        self.mcdata = mcdata
        self.batch_mcdata = list(split_generator(mcdata, batch))
        self.batch = batch
        if "weight" in mcdata:
            mc_weight = tf.convert_to_tensor(mcdata["weight"], dtype="float64")
            self.mc_weight = mc_weight / tf.reduce_sum(mc_weight)
        else:
            self.mc_weight = tf.convert_to_tensor(
                [1 / n_mcdata] * n_mcdata, dtype="float64"
            )
        self.batch_mc_weight = list(data_split(self.mc_weight, self.batch))
        self.gauss_constr = GaussianConstr(self.vm, gauss_constr)
        self.cached_mc = {}

    def get_params(self, trainable_only=False):
        return self.vm.get_all_dic(trainable_only)

    def get_nll(self, x={}):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        """
        self.model.set_params(x)
        if type(self.model) == Model_new:
            nll, g = self.get_nll_grad(x)
        else:
            nll = self.model.nll(
                self.data,
                self.mcdata,
                weight=self.weight,
                mc_weight=self.mc_weight,
            )
            self.n_call += 1
        return nll

    def __call__(self, x={}):
        self.cached_nll = (
            self.get_nll(x) + self.gauss_constr.get_constrain_term()
        )
        return self.cached_nll

    def get_grad(self, x={}):
        """
        :param x: List. Values of variables.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        nll, g = self.get_nll_grad(x)
        return g

    def grad(self, x={}):
        return self.get_grad(x) + self.gauss_constr.get_constrain_grad()

    def get_nll_grad(self, x={}):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        self.model.set_params(x)
        nll, g = self.model.nll_grad_batch(
            self.batch_data,
            self.batch_mcdata,
            weight=list(data_split(self.weight, self.batch)),
            mc_weight=self.batch_mc_weight,
        )
        self.n_call += 1
        return nll, g

    @time_print
    def nll_grad(self, x={}):
        nll, g = self.get_nll_grad(x)
        constr = self.gauss_constr.get_constrain_term()
        constr_grad = self.gauss_constr.get_constrain_grad()
        self.cached_nll = nll + constr
        return float(self.cached_nll), g + constr_grad

    def get_nll_grad_hessian(self, x={}, batch=None):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        if batch is None:
            batch = self.batch
        self.model.set_params(x)
        if type(self.model) == Model_new:
            nll, g, h = self.model.nll_grad_hessian(
                self.batch_data,
                self.batch_mcdata,
                weight=list(data_split(self.weight, self.batch)),
                mc_weight=data_split(self.mc_weight, self.batch),
            )
        else:
            nll, g, h = self.model.nll_grad_hessian(
                self.data,
                self.mcdata,
                weight=self.weight,
                batch=batch,
                mc_weight=self.mc_weight,
            )
        return nll, g, h

    def nll_grad_hessian(self, x={}, batch=None):
        if batch is None:
            batch = self.batch
        nll, g, h = self.get_nll_grad_hessian(x, batch)
        constr = self.gauss_constr.get_constrain_term()
        constr_grad = self.gauss_constr.get_constrain_grad()
        constr_hessian = self.gauss_constr.get_constrain_hessian()
        return nll + constr, g + constr_grad, h + constr_hessian

    @time_print
    def grad_hessp(self, x, p, batch=None):
        if batch is None:
            batch = self.batch
        g, h = self.get_grad_hessp(x, p, batch)
        constr_grad = self.gauss_constr.get_constrain_grad()
        constr_hessian = 0.0  # self.gauss_constr.get_constrain_hessp(p)
        return g + constr_grad, h + constr_hessian

    def get_grad_hessp(self, x, p, batch):
        self.model.set_params(x)
        grad, hessp = self.model.grad_hessp_batch(
            p,
            self.batch_data,
            self.batch_mcdata,
            weight=list(data_split(self.weight, self.batch)),
            mc_weight=self.batch_mc_weight,
        )
        return grad, hessp


class CombineFCN(object):
    """
    This class implements methods to calculate the NLL as well as its derivatives for a general function.

    :param model: List of model object.
    :param data: List of data array.
    :param mcdata: list of MCdata array.
    :param bg: list of Background array.
    :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
    """

    def __init__(
        self,
        model=None,
        data=None,
        mcdata=None,
        bg=None,
        fcns=None,
        batch=65000,
        gauss_constr={},
    ):
        if fcns is None:
            assert model is not None, "model required"
            assert data is not None, "data required"
            assert mcdata is not None, "mcdata required"
            self.fcns = []
            self.cached_nll = 0.0
            if bg is None:
                bg = _loop_generator(None)
            for model_i, data_i, mcdata_i, bg_i in zip(
                model, data, mcdata, bg
            ):
                self.fcns.append(FCN(model_i, data_i, mcdata_i, bg_i))
        else:
            self.fcns = list(fcns)
        self.vm = self.fcns[0].vm
        self.gauss_constr = GaussianConstr(self.vm, gauss_constr)

    def get_params(self, trainable_only=False):
        return self.vm.get_all_dic(trainable_only)

    def get_nll(self, x={}):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        """
        nlls = []
        for i in self.fcns:
            nlls.append(i.get_nll(x))
        return sum(nlls)

    def __call__(self, x={}):
        self.cached_nll = (
            self.get_nll(x) + self.gauss_constr.get_constrain_term()
        )
        return self.cached_nll

    def get_grad(self, x={}):
        """
        :param x: List. Values of variables.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        gs = []
        for i in self.fcns:
            g = i.get_grad(x)
            gs.append(g)
        return sum(gs)

    def grad(self, x={}):
        return self.get_grad(x) + self.gauss_constr.get_constrain_grad()

    def get_nll_grad(self, x={}):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        nlls = []
        gs = []
        for i in self.fcns:
            nll, g = i.get_nll_grad(x)
            nlls.append(nll)
            gs.append(g)
        return sum(nlls), tf.reduce_sum(gs, axis=0)

    @time_print
    def nll_grad(self, x={}):
        nll, g = self.get_nll_grad(x)
        constr = self.gauss_constr.get_constrain_term()
        constr_grad = self.gauss_constr.get_constrain_grad()
        self.cached_nll = nll + constr
        return float(self.cached_nll), g + constr_grad

    def get_nll_grad_hessian(self, x={}, batch=None):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        nlls = []
        gs = []
        hs = []
        for i in self.fcns:
            nll, g, h = i.get_nll_grad_hessian(x)
            nlls.append(nll)
            gs.append(g)
            hs.append(h)
        # print("NLL list: ",nlls)
        # print("Gradient List: ",tf.transpose(gs))
        return (
            tf.reduce_sum(nlls, axis=0),
            tf.reduce_sum(gs, axis=0),
            tf.reduce_sum(hs, axis=0),
        )

    def nll_grad_hessian(self, x={}, batch=None):
        nll, g, h = self.get_nll_grad_hessian(x, batch)
        constr = self.gauss_constr.get_constrain_term()
        constr_grad = self.gauss_constr.get_constrain_grad()
        constr_hessian = self.gauss_constr.get_constrain_hessian()
        return nll + constr, g + constr_grad, h + constr_hessian

    def get_grad_hessp(self, x, p, batch):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        hs = []
        gs = []
        for i in self.fcns:
            g, h = i.get_grad_hessp(x, p, batch)
            hs.append(h)
            gs.append(g)
        return tf.reduce_sum(gs, axis=0), tf.reduce_sum(hs, axis=0)

    def grad_hessp(self, x, p, batch=None):
        grad, hessp = self.get_grad_hessp(x, p, batch)
        constr_grad = self.gauss_constr.get_constrain_grad()
        return grad + constr_grad, hessp


class MixLogLikehoodFCN(CombineFCN):

    """
    This class implements methods to calculate the NLL as well as its derivatives for a general function.

    :param model: List of model object.
    :param data: List of data array.
    :param mcdata: list of MCdata array.
    :param bg: list of Background array.
    :param batch: The length of array to calculate as a vector at a time. How to fold the data array may depend on the GPU computability.
    """

    def __init__(
        self,
        model,
        data,
        mcdata,
        bg=None,
        batch=65000,
        gauss_constr={},
    ):
        self.cached_nll = 0.0
        assert model is not None, "model required"
        assert data is not None, "data required"
        assert mcdata is not None, "mcdata required"
        self.fcns = []
        self.cached_nll = 0.0
        if bg is None:
            bg = _loop_generator(None)
        self.datas = []
        self.weight_phsps = []
        self.n_datas = []
        self.model = model
        for model_i, data_i, mcdata_i, bg_i in zip(model, data, mcdata, bg):
            data_s = model_i.mix_data_bakcground(data_i, bg_i)
            self.datas.append(data_s)
            weight_phsp = type(mcdata_i)(
                {k: v for k, v in mcdata_i.items()}
            )  # simple copy
            w = weight_phsp.get_weight()
            weight_phsp["weight"] = w / tf.reduce_sum(w)
            self.n_datas.append(tf.reduce_sum(data_s.get_weight()))
            self.weight_phsps.append(list(split_generator(weight_phsp, batch)))
            self.fcns.append(FCN(model_i, data_i, mcdata_i, bg_i))
        self.data_merge = list(split_generator(data_merge(*self.datas), batch))
        self.vm = self.model[0].vm
        self.gauss_constr = GaussianConstr(self.vm, gauss_constr)

    def get_nll_grad(self, x={}):
        """
        :param x: List. Values of variables.
        :return nll: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        """
        [i.set_params(x) for i in self.model]
        nlls = []
        gs = []
        nll, g = self.model[0].sum_nll_grad_bacth(self.data_merge)
        nlls.append(nll)
        gs.append(g)
        for i, k, l in zip(self.model, self.weight_phsps, self.n_datas):
            nll, g = i.sum_log_integral_grad_batch(k, l)
            nlls.append(nll)
            gs.append(g)
        print(sum(nlls))
        return sum(nlls), tf.reduce_sum(gs, axis=0)



import numpy as np
import tensorflow as tf

from tf_pwa.amp import variable_scope
from tf_pwa.config import set_config
from tf_pwa.function import nll_funciton
from tf_pwa.variable import Variable, VarsManager, combineVM


def test_variable():
    with variable_scope() as vm:
        m = Variable("R_m", value=2.1)  # trainable is True by default
        g_ls = Variable("A2BR_H", shape=[3], cplx=True)
        fcr = Variable("R_total", cplx=True)
        m1 = Variable("R1_m", value=2.3)
        g_ls1 = Variable("A2BR1_H", shape=[3], cplx=True)
        fcr1 = Variable("R1_total", cplx=True)

        print(g_ls.value)
        print(g_ls())

        m.fixed(2.4)
        g_ls.sameas(g_ls1)
        fcr.r_shareto(fcr1)


def test_variable2():
    set_config("polar", False)
    with variable_scope() as vm:
        b = Variable("R_total", cplx=True)
        assert vm.complex_vars["R_total"] == False
    set_config("polar", True)
    with variable_scope() as vm:
        b = Variable("R_total", cplx=True)
        assert vm.complex_vars["R_total"] == True


def test_combine_vm():
    a = VarsManager("a")
    a_v = Variable("a", vm=a)
    b = VarsManager("b")
    b_v = Variable("b", vm=b)
    com_vm = combineVM(a, b)
    com_vm.set("aa", 3.0)
    assert float(a_v()) == 3.0


def test_minimize():
    with variable_scope() as vm:
        m = Variable("R_m", value=2.1)
        vm.set_bound({"R_m": [-2, 3]})

        def f():
            return m() * m()

        ret = vm.minimize(f)
        print(ret)
        assert np.allclose(m().numpy(), 0.0)


def test_minimize2():
    data = np.linspace(-1.5, 1.5, 10000)
    w = tf.cos(data) + 1
    np.random.seed(2)
    cut = np.random.random(data.shape) * np.max(w) * 1.1 < w
    data = data[cut]

    phsp = np.linspace(-1.5, 1.5, 10000)

    with variable_scope() as vm:
        a = Variable("R_m", cplx=True)
        vm.set_bound({"R_mr": (-2, None)})

        def f(x):
            fx = tf.cos(x) + tf.abs(a())
            return fx

        ret = vm.minimize(nll_funciton(f, data, phsp))
        print(ret)
        assert abs(float(np.abs(a())) - 1.0) < 0.2


def test_minimize():
    with variable_scope() as vm:
        m = Variable("R_m", value=2.1)
        vm.set_bound({"R_m": (None, 3)})

        def f():
            return m() * m()

        from scipy.optimize import minimize as mini

        ret = vm.minimize(
            f, method=lambda g, x: mini(g, x, jac=True, method="L-BFGS-B")
        )
        print(ret)
        assert abs(m().numpy()) < 1e-6


def test_polar():
    with variable_scope() as vm:
        a = Variable("a", cplx=True)
        b = a().numpy()
        vm.rp2xy_all()
        c = a().numpy()
        vm.xy2rp_all()
        d = a().numpy()
        vm.std_polar_all()
        e = vm.get("ai")
        vm.trans_params(True)
        f = a().numpy()
        assert np.allclose(b, c)
        assert np.allclose(c, d)
        assert np.allclose(d, f)
        assert -3.2 < e and e < 3.2


def test_refresh_vars():
    with variable_scope() as vm:
        Variable("a", cplx=True)
        Variable("b", value=1.0)
        Variable("c", value=1.0, range_=[0, 3])

        Variable("b", value=1.0)
        Variable("d", value=1.0, fix=True)
        vm.refresh_vars()


def test_rename():
    with variable_scope() as vm:
        Variable("a", value=1.0)
        vm.rename_var("a", "b")
        assert vm.get("b") == 1.0
        assert "a" not in vm.variables
        a = Variable("d", cplx=True)  # BUG: cannot use "a"
        a.set_value([2, 3])
        vm.rename_var("d", "c", True)
        assert vm.get("cr") == 2
        assert vm.get("ci") == 3



"""
This module provides some functions that may be useful in other modules.
"""
import functools
import json
import math
import time
import warnings

import numpy as np


class AttrDict(dict):
    __setattr__ = dict.__setitem__
    __getattr__ = dict.__getitem__


has_yaml = True
try:
    import yaml
except ImportError:
    has_yaml = False


def _load_json_file(name):
    with open(name) as f:
        return json.load(f)


def _load_yaml_file(name):
    with open(name) as f:
        return yaml.load(f, Loader=yaml.FullLoader)


def load_config_file(name):
    """
    Load config file such as **Resonances.yml**.

    :param name: File name. Either yml file or json file.
    :return: Dictionary read from the file.
    """
    if name.endswith("json"):
        return _load_json_file(name)
    if has_yaml:
        if name.endswith("yml"):
            return _load_yaml_file(name)
        return _load_yaml_file(name + ".yml")
    else:
        print("No yaml support, using json file")
        return _load_json_file(name + ".json")


def flatten_dict_data(data, fun="{}/{}".format):
    """
    Flatten nested dictionary data into one layer dictionary.

    :return: Dictionary
    """
    if isinstance(data, dict):
        ret = {}
        for i in data:
            tmp = flatten_dict_data(data[i])
            if isinstance(tmp, dict):
                for j in tmp:
                    ret[fun(i, j)] = tmp[j]
            else:
                ret[i] = tmp
        return ret
    else:
        return data


flatten_np_data = lambda data: flatten_dict_data(
    data, fun=lambda x, y: "{}{}".format(y, x[3:])
)


def error_print(x, err=None):
    """
    It returns a format string "value +/- error". The precision is modified according to ``err``

    :param x: Value
    :param err: Error
    :return: String
    """
    if err is None:
        return ("{}").format(x)
    if err <= 0 or math.isnan(err):
        return ("{} ? {}").format(x, err)
    d = math.ceil(math.log10(err))
    b = 10 ** d
    b_err = err / b
    b_val = x / b
    if b_err < 0.355:  # 0.100 ~ 0.354
        dig = 2
    elif b_err < 0.950:  # 0.355 ~ 0.949
        dig = 1
    else:  # 0.950 ~ 0.999
        dig = 0
    err = round(b_err, dig) * b
    x = round(b_val, dig) * b
    d_p = dig - d
    if d_p > 0:
        return ("{0:.%df} +/- {1:.%df}" % (d_p, d_p)).format(x, err)
    return ("{0:.0f} +/- {1:.0f}").format(x, err)


def pprint(dicts):
    """
    Print dictionary using json format.
    """
    try:
        s = json.dumps(dicts, indent=2)
        print(s, flush=True)
    except:
        print(dicts, flush=True)


def print_dic(dic):
    """
    Another way to print dictionary.
    """
    if type(dic) == dict:
        for i in dic:
            print(i + " :\t", dic[i])
    else:
        print(dic)


def std_polar(rho, phi):
    """
    To standardize a polar variable. By standard form, it means :math:`\\rho>0, -\\pi<\\phi<\\pi`.

    :param rho: Real number
    :param phi: Real number
    :return: ``rho``, ``phi``
    """
    if rho < 0:
        rho = -rho
        phi += math.pi
    while phi < -math.pi:
        phi += 2 * math.pi
    while phi > math.pi:
        phi -= 2 * math.pi
    return rho, phi


def deep_iter(base, deep=1):
    for i in base:
        if deep == 1:
            yield [i]
        else:
            for j in deep_iter(base, deep - 1):
                yield [i] + j


def deep_ordered_iter(base, deep=1):
    ids = list(base)
    size = len(ids)
    for i in deep_ordered_range(size, deep):
        yield [ids[j] for j in i]


def deep_ordered_range(size, deep=1, start=0):
    for i in range(start, size):
        if deep <= 1:
            yield [i]
        else:
            for j in deep_ordered_range(size, deep - 1, i + 1):
                yield [i] + j


# from amplitude.py
def is_complex(x):
    """
    If **x** is of type ``complex``, it returns ``True``.
    """
    try:
        y = complex(x)
    except:
        return False
    return True


# from model.py
def array_split(data, batch=None):
    """Split a data array. **batch** is the number of data in a row."""
    if batch is None:
        return [data]
    ret = []
    n_data = data[0].shape[0]
    n_split = (n_data + batch - 1) // batch
    for i in range(n_split):
        tmp = []
        for data_i in data:
            tmp.append(data_i[i * batch : min(i * batch + batch, n_data)])
        ret.append(tmp)
    return ret


def time_print(f):
    """It provides a wrapper to print the time cost on a process."""

    @functools.wraps(f)
    def g(*args, **kwargs):
        now = time.time()
        ret = f(*args, **kwargs)
        print(f.__name__, " cost time:", time.time() - now)
        return ret

    return g


def std_periodic_var(p, mid=0.0, pi=math.pi):
    """
    Transform a periodic variable into its range.

    >>> std_periodic_var(math.pi)
    -3.1415...

    >>> std_periodic_var(2*math.pi + 0.01)
    0.0...

    :param p: Value
    :param mid: The middle value
    :param pi: Half-range
    :return: The transformed value
    """
    twopi = 2 * pi
    while p < mid - pi:
        p += twopi
    while p >= mid + pi:
        p -= twopi
    return p


def check_positive_definite(m):
    """check if matrix m is postive definite

    >>> check_positive_definite([[1.0,0.0],[0.0, 0.1]])
    True

    >>> check_positive_definite([[1.0,0.0],[1.0,-0.1]])
    eigvalues:  [-0.1  1. ]
    False

    """
    e, v = np.linalg.eig(m)
    if np.all(e > 0.0):
        return True
    warnings.warn("matrix is not positive definited")
    print("eigvalues: ", e)
    return False


def tuple_table(fit_frac):
    names = []
    for i in fit_frac:
        if isinstance(i, str):
            names.append(i)
    n_items = len(names)
    table = [[None] * (n_items + 1) for i in range(n_items + 1)]

    for k, v in fit_frac.items():
        if isinstance(k, tuple):
            a, b = k
            table[names.index(a) + 1][names.index(b) + 1] = v
        else:
            a, b = k, k
            table[names.index(a) + 1][names.index(b) + 1] = v

    for i, name in enumerate(names):
        table[i + 1][0] = name
        table[0][i + 1] = name

    return table


def fit_normal(data, weights=None):
    """
    Fit data distribution with Gaussian distribution. Though minimize the negative log likelihood function

    .. math::
        - \\ln L = \\frac{1}{2}\\sum w_i \\frac{(\\mu - x_i )^2}{\\sigma^2} + (\\sum w_i) \\ln (\\sqrt{2\pi} \\sigma )

    the fit result can be solved as

    .. math::
        \\frac{\\partial (-\\ln L)}{\\partial \\mu} = 0 \\Rightarrow \\bar{\\mu} = \\frac{\\sum w_i x_i}{ \\sigma^2 \\sum w_i}

    .. math::
        \\frac{\\partial (-\\ln L)}{\\partial \\sigma} = 0 \\Rightarrow \\bar{\\sigma} = \\sqrt{\\frac{\\sum w_i (\\bar{\\mu} - x_i)^2}{\\sum w_i}}

    From hessian

    .. math::
        \\frac{\\partial^2 (-\\ln L)}{\\partial \\mu^2} = \\frac{\\sum w_i}{\\sigma^2}

    .. math::
        \\frac{\\partial^2 (-\\ln L)}{\\partial \\sigma^2} = 3\\sum \\frac{\\sum w_i (\\mu - x)^2}{\\sigma^4} - \\frac{\\sum w_i}{\\sigma^2}

    the error matrix can wrotten as  [[ :math:`\\bar{\\sigma}^2/N` , 0], [0, :math:`\\bar{\\sigma}^2/(2N)` ]] .

    """
    if weights is None:
        weights = np.ones_like(data)
    else:
        weights = np.sum(weights) / np.sum(weights ** 2) * weights
    N = np.sum(weights)
    mu = np.sum(weights * data) / N
    sigma = np.sqrt(np.sum(weights * (data - mu) ** 2) / N)
    mu_error = sigma / np.sqrt(N)
    sigma_error = mu_error / np.sqrt(2)
    return np.array([mu, sigma]), np.array([mu_error, sigma_error])



import os.path
import sys

import numpy as np

from tf_pwa.applications import cal_hesse_error, corr_coef_matrix

# import tf_pwa
from tf_pwa.config_loader import ConfigLoader
from tf_pwa.utils import error_print

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def main():
    """Calculate errors of a given set of parameters and their correlation coefficients."""
    import argparse

    parser = argparse.ArgumentParser(
        description="calculate errors of a given set of parameters and their correlation coefficients"
    )
    parser.add_argument(
        "--params", default="final_params.json", dest="params_file"
    )
    results = parser.parse_args()

    cal_errors(params_file=results.params_file)


def cal_errors(params_file):
    config = ConfigLoader("config.yml")
    config.set_params(params_file)
    fcn = config.get_fcn()
    fcn.model.Amp.vm.rp2xy_all()  # we can use this to transform all complex parameters to xy-coordinates, since the Hesse errors of xy are more statistically reliable
    params = config.get_params()
    errors, config.inv_he = cal_hesse_error(
        fcn, params, check_posi_def=True, save_npy=True
    )  # obtain the Hesse errors and the error matrix (inverse Hessian)
    print("\n########## fit parameters in XY-coordinates:")
    errors = dict(zip(fcn.model.Amp.vm.trainable_vars, errors))
    for key, value in config.get_params().items():
        print(key, error_print(value, errors.get(key, None)))

    print("\n########## correlation matrix:")
    print("Matrix index:\n", fcn.model.Amp.vm.trainable_vars)
    print(
        "Correlation Coefficients:\n", corr_coef_matrix(config.inv_he)
    )  # obtain the correlation matrix using the inverse Hessian


if __name__ == "__main__":
    main()



"""
Scripts for calculating fit fraction of each LS component.

Need config.yml and final_params.json (and error_matrix.npy for error).

"""

import itertools

import numpy as np
import tensorflow as tf

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import split_generator
from tf_pwa.experimental.opt_int import split_gls
from tf_pwa.model.model import sum_gradient


def cal_each_ls(var, dg, dec_chain, data, weights):
    """
    calculate amplitude square for each L-S coupling.
    """
    cached = []
    for i, dc in split_gls(dec_chain):
        int_mc, grad = sum_gradient(dg.sum_amp, data, var, weights)
        cached.append((int_mc, grad))
    return cached


def cal_each_decay_chain(var, dec, data, weights):
    """
    calculate amplitude square for each decay chain (resonant).
    """
    hij = []
    used_chains = dec.chains_idx
    index = []
    for k, i in enumerate(dec):
        dec.set_used_chains([k])
        tmp = cal_each_ls(var, dec, i, data, weights)
        hij.append(tmp)
        index.append(i)
    dec.set_used_chains(used_chains)  # recover
    return index, hij


def cal_frac(amp, phsp, err_matrix, batch=300000):
    decay_group = amp.decay_group
    var = amp.trainable_variables

    # split large data to avoid OOM
    phsp = list(split_generator(phsp, batch))
    weights = [i.get("weight", 1.0) for i in phsp]

    # denominator part
    int_mc, grad_total = sum_gradient(amp, phsp, var, weights)
    int_mc = int_mc.numpy()
    grad_total = np.array([i.numpy() for i in grad_total])

    idx, amp_matrix2 = cal_each_decay_chain(var, decay_group, phsp, weights)

    # numerator part for each decay chain
    total = 0
    for k, v in zip(idx, amp_matrix2):
        print("decay chain:", k)
        ls_list = [i.get_ls_list() for i in k]
        ls = itertools.product(*ls_list)
        for i, (value, grad) in enumerate(v):
            # fraction gradient
            value = value.numpy()
            grad = np.array([i.numpy() for i in grad])
            grad_f = (grad - grad_total * value / int_mc) / int_mc

            # error propagation
            error = np.sqrt(np.dot(np.dot(err_matrix, grad_f), grad_f))
            print(
                f"  frac of ls: {next(ls)} is "
                f"{value/int_mc:.4f} +/- {error:.4f}"
            )
            total += value
    print("sum diagonal:", total / int_mc)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="calculate fit fractions")
    parser.add_argument("-c", "--config", default="config.yml")
    parser.add_argument("-i", "--init_params", default="final_params.json")
    parser.add_argument("-e", "--error_matrix", default="error_matrix.npy")
    results = parser.parse_args()

    # load model and parameters and error matrix
    config = ConfigLoader(results.config)
    config.set_params(results.init_params)
    err_matrix = np.load(results.error_matrix)

    amp = config.get_amplitude()
    phsp = config.get_phsp_noeff()  # get_data("phsp")[0]

    cal_frac(amp, phsp, err_matrix)


if __name__ == "__main__":
    main()



import json

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as signal
import yaml
from mpl_toolkits.mplot3d.axes3d import Axes3D
from scipy.interpolate import interp1d

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.experimental.extra_amp import spline_matrix

# import mplhep
# plt.style.use(mplhep.style.LHCb)


def vialid_name(s):
    return s.replace("+", ".")


def polar_err(r, phi, r_e, phi_e):
    """polar errors for r and phi"""
    # print(r, phi, r_e, phi_e)
    dxdr = np.cos(phi)
    dxdphi = r * np.sin(phi)
    dydr = np.sin(phi)
    dydphi = -r * np.cos(phi)
    x_e = np.sqrt((dxdr * r_e) ** 2 + (dxdphi * phi_e) ** 2)
    y_e = np.sqrt((dydr * r_e) ** 2 + (dydphi * phi_e) ** 2)
    # print(x_e, y_e)
    return x_e, y_e


def dalitz_weight(s12, m0, m1, m2, m3):
    """phase space weight in dalitz plot"""
    m12 = np.sqrt(s12)
    m12 = np.where(m12 > (m1 + m2), m12, m1 + m2)
    m12 = np.where(m12 < (m0 - m3), m12, m0 - m3)
    # if(mz < (m_d+m_pi)) return 0;
    # if(mz > (m_b-m_pi)) return 0;
    E2st = 0.5 * (m12 * m12 - m1 * m1 + m2 * m2) / m12
    E3st = 0.5 * (m0 * m0 - m12 * m12 - m3 * m3) / m12

    p2st2 = E2st * E2st - m2 * m2
    p3st2 = E3st * E3st - m3 * m3
    p2st = np.sqrt(np.where(p2st2 > 0, p2st2, 0))
    p3st = np.sqrt(np.where(p3st2 > 0, p3st2, 0))
    return p2st * p3st


def load_params(
    config_file="config.yml", params="final_params.json", res="li(1+)S"
):
    with open(params) as f:
        final_params = json.load(f)
    val = final_params["value"]
    err = final_params["error"]
    with open(config_file) as f:
        config = yaml.safe_load(f)
    xi = config["particle"][res].get("points", None)
    if xi is None:
        m_max = config["particle"][res].get("m_max", None)
        m_min = config["particle"][res].get("m_min", None)
        N = config["particle"][res].get("interp_N", None)
        dx = (m_max - m_min) / (N - 1)
        xi = [m_min + dx * i for i in range(N)]
    N = len(xi)
    head = "{}_point".format(vialid_name(res))
    r = np.array(
        [0] + [val["{}_{}r".format(head, i)] for i in range(N - 2)] + [0]
    )
    phi = np.array(
        [0] + [val["{}_{}i".format(head, i)] for i in range(N - 2)] + [0]
    )
    r_e = np.array(
        [0, 0]
        + [
            err.get("{}_{}r".format(head, i), r[i] * 0.1)
            for i in range(1, N - 2)
        ]
        + [0]
    )
    phi_e = np.array(
        [0, 0]
        + [
            err.get("{}_{}i".format(head, i), phi[i] * 0.1)
            for i in range(1, N - 2)
        ]
        + [0]
    )
    return np.array(xi), r, phi, r_e, phi_e


def trans_r2xy(r, phi, r_e, phi_e):
    """r,phi -> x,y """
    x = np.array(r) * np.cos(phi)
    y = np.array(r) * np.sin(phi)
    err = np.array(
        [polar_err(i, j, k, l) for i, j, k, l in zip(r, phi, r_e, phi_e)]
    )
    return x, y, err[:, 0], err[:, 1]


def plot_x_y(name, x, y, x_i, y_i, xlabel, ylabel, ylim=(None, None)):
    """plot x vs y"""
    plt.clf()
    plt.plot(x, y)
    plt.scatter(x_i, y_i)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.ylim(ylim)
    plt.savefig(name)


def plot_phi(name, m, phi, m_i, phi_i):
    """ plot phi and gradient of phi"""
    grad = phi[2:] - phi[:-2]
    mask = (phi < 3) & (phi > -3)
    grad_max = np.mean(np.abs(grad))
    # grad_max = np.max(grad[mask[1:-1]])

    (idx,) = signal.argrelextrema(grad, np.less)

    plt.clf()

    # plt.plot(m, pq/np.max(pq))# np.sqrt(x_new**2+y_new**2)**2)
    plt.plot(m[1:-1], grad / grad_max, label="$\\Delta \\phi$ ")
    plt.plot(m, phi, label="$\\phi$")  # np.sqrt(x_new**2+y_new**2)**2)

    m_delta = m[idx + 1]
    print("min Delta phi in mass:", m_delta)
    plt.scatter(m_delta, [-np.pi] * len(m_delta))
    plt.scatter(m_i, phi_i, label="points")
    plt.xlabel("mass")
    plt.ylabel("$\\phi$")
    plt.ylim((-np.pi, np.pi))
    plt.legend()
    plt.savefig(name)


def plot_x_y_err(name, x, y, x_e, y_e):
    """plot eror bar of x y"""
    plt.clf()
    plt.errorbar(x, y, xerr=x_e, yerr=y_e)
    plt.xlabel("real R(m)")
    plt.ylabel("imag R(m)")
    plt.savefig(name)


def plot3d_m_x_y(name, m, x, y):
    fig = plt.figure()
    axes3d = Axes3D(fig)
    axes3d.plot(m, x, y)
    axes3d.set_xlabel("m")
    axes3d.set_ylabel("real R(m)")
    axes3d.set_zlabel("imag R(m)")

    def update(frame):
        axes3d.view_init(elev=30, azim=frame)
        return None

    anim = animation.FuncAnimation(
        fig, update, interval=10, frames=range(0, 360, 10)
    )
    anim.save(name, writer="imagemagick")


def plot_all(
    res="MI(1+)S",
    config_file="config.yml",
    params="final_params.json",
    prefix="figure/",
):
    """plot all figure"""
    config = ConfigLoader(config_file)
    config.set_params(params)
    particle = config.get_decay().get_particle(res)

    mi, r, phi_i, r_e, phi_e = load_params(config_file, params, res)
    x, y, x_e, y_e = trans_r2xy(r, phi_i, r_e, phi_e)

    m = np.linspace(mi[0], mi[-1], 1000)
    M_Kpm = 0.49368
    M_Dpm = 1.86961
    M_Dstar0 = 2.00685
    M_Bpm = 5.27926
    # x_new = interp1d(xi, x, "cubic")(m)
    # y_new = interp1d(xi, y, "cubic")(m)
    rm_new = particle.interp(m).numpy()
    x_new, y_new = rm_new.real, rm_new.imag

    pq = dalitz_weight(m * m, M_Bpm, M_Dstar0, M_Dpm, M_Kpm)
    pq_i = dalitz_weight(mi * mi, M_Bpm, M_Dstar0, M_Dpm, M_Kpm)

    phi = np.arctan2(y_new, x_new)
    r2 = x_new * x_new + y_new * y_new

    plot_phi(f"{prefix}phi.png", m, phi, mi, np.arctan2(y, x))
    plot_x_y(
        f"{prefix}r2.png",
        m,
        r2,
        mi,
        r * r,
        "mass",
        "$|R(m)|^2$",
        ylim=(0, None),
    )
    plot_x_y(f"{prefix}x_y.png", x_new, y_new, x, y, "real R(m)", "imag R(m)")
    plot_x_y_err(
        f"{prefix}x_y_err.png", x[1:-1], y[1:-1], x_e[1:-1], y_e[1:-1]
    )
    plot_x_y(
        f"{prefix}r2_pq.png",
        m,
        r2 * pq,
        mi,
        r * r * pq_i,
        "mass",
        "$|R(m)|^2 p \cdot q$",
        ylim=(0, None),
    )
    plot3d_m_x_y(f"{prefix}m_r.gif", m, x_new, y_new)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="plot interpolation")
    parser.add_argument("particle", type=str)
    parser.add_argument("-c", "--config", default="config.yml", dest="config")
    parser.add_argument(
        "-i", "--params", default="final_params.json", dest="params"
    )
    parser.add_argument("-p", "--prefix", default="figure/", dest="prefix")
    results = parser.parse_args()
    plot_all(results.particle, results.config, results.params, results.prefix)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# from tf_pwa.phasespace import  PhaseSpaceGenerator
import itertools
import os.path
import sys

from tf_pwa.amp import DecayChain, get_decay, get_particle

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def jp_seq(max_J=2):
    """all JP combination"""
    for j in range(max_J + 1):
        for p in [-1, 1]:
            yield j, p


def possible_jp(decay_chain, max_J=2):
    """get possible resonances jp of J <= max_J"""
    ret = []
    A = decay_chain.top
    outs = decay_chain.outs
    jp_list = [jp_seq(max_J) for _ in decay_chain.inner]

    ret = {}
    for jps in itertools.product(*jp_list):
        res_map = {}
        for jp, res in zip(jps, decay_chain.inner):
            j, p = jp
            res_n = get_particle("{}_{}{}".format(res, j, p), J=j, P=p)
            res_map[res] = res_n
        possible_ls = []
        for i in decay_chain:
            core = res_map.get(i.core, i.core)
            outs = [res_map.get(j, j) for j in i.outs]
            dec = get_decay(core, outs, p_break=i.p_break, disable=True)
            # print(core.J, core.P, outs[0].J, outs[0].P, outs[1].J, outs[1].P)
            # print(dec.get_ls_list())
            possible_ls.append(dec.get_ls_list())
        if min([len(i) for i in possible_ls]) > 0:
            ret[jps] = dict(zip(decay_chain, possible_ls))
    return ret


def main():
    a = get_particle("A", J=0, P=-1)
    b = get_particle("B", J=1, P=-1)
    c = get_particle("C", J=0, P=-1)
    d = get_particle("D", J=0, P=-1)
    r = get_particle("R")

    dec1 = get_decay(a, [r, d], p_break=True)
    dec2 = get_decay(r, [b, c])

    decs = DecayChain([dec1, dec2])

    # chains = DecayChain.from_particles(a, [b, c, d])
    # for i in chains:
    jp_list = possible_jp(decs)
    for jps in jp_list:
        sign = lambda x: "+" if x == 1 else "-"
        jp_display = ["{}{}".format(j, sign(p)) for j, p in jps]
        print("Resonances JP: ", dict(zip(decs.inner, jp_display)))
        print("  Possible (l, s): ", jp_list[jps])


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import json
import os.path
import sys
from itertools import combinations
from pprint import pprint

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

from tf_pwa.config_loader import ConfigLoader, MultiConfig
from tf_pwa.data import data_index
from tf_pwa.utils import error_print

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def get_data(config_file="config.yml", init_params="init_params.json"):

    config = ConfigLoader(config_file)
    try:
        config.set_params(init_params)
        print("using {}".format(init_params))
    except Exception as e:
        print("using RANDOM parameters")

    phsp = config.get_data("phsp")

    for i in config.full_decay:
        print(i)
        for j in i:
            print(j.get_ls_list())

    print("\n########### initial parameters")
    print(json.dumps(config.get_params(), indent=2))
    params = config.get_params()

    amp = config.get_amplitude()
    pw = amp.partial_weight(phsp)
    pw_if = amp.partial_weight_interference(phsp)
    weight = amp(phsp)
    print(weight)
    return config, amp, phsp, weight, pw, pw_if


def get_params():
    with open("test_results.json") as f:
        data = json.load(f)

    data_infer = [i["NR(1+)DxDs1(2700)"] for i in data["fracs"]]
    data_0 = [i["Ds1(2700)"] for i in data["fracs"]]
    data_1 = [i["NR(1+)S"] for i in data["fracs"]]
    data_2 = [i["NR(1+)D"] for i in data["fracs"]]

    idx = np.argmax(
        np.fabs(np.array(data_0) * np.array(data_1) * np.array(data_2))
    )
    print(idx, data_infer[idx])
    return data["prarms"][idx]


def test_plot(init_params):

    import matplotlib

    matplotlib.rcParams["figure.figsize"] = [10, 10]  # for square canvas
    matplotlib.rcParams["figure.subplot.left"] = 0.05
    matplotlib.rcParams["figure.subplot.bottom"] = 0.05
    matplotlib.rcParams["figure.subplot.right"] = 0.95
    matplotlib.rcParams["figure.subplot.top"] = 0.95
    matplotlib.rcParams["font.size"] = 5
    config, amp, data, weight, pw, pw_if = get_data(
        init_params=init_params
    )  # init_params=get_params())

    m_DsDst = data_index(
        data, config.plot_params.get_data_index("mass", "R_BC")
    )
    m_DsK = data_index(data, config.plot_params.get_data_index("mass", "R_BD"))
    m_DstK = data_index(
        data, config.plot_params.get_data_index("mass", "R_CD")
    )

    val = m_DsDst

    sw = np.sum(weight) / weight.shape[0]
    print(sw)
    # plt.hist(val, weights=[sw]*weight.shape[0], bins=100, label="PHSP")
    # plt.hist(val, weights=weight, bins=100, label="data")
    label = amp.chains_particle()
    label = [str(i[0]) for i in label]  #
    n_label = len(label)
    for i, w in zip(label, pw):
        print(i, np.sum(w) / sw)

    if n_label % 2 == 0:
        a, b = n_label // 2, n_label - 1
    else:
        a, b = (n_label - 1) // 2, n_label
    fig = plt.figure(figsize=(4, 3))
    for i, j in enumerate(combinations(range(n_label), 2)):
        # ax = plt.subplot(a,b,i+1)
        plt.clf()
        ax = plt.subplot(1, 1, 1)
        ax.hist(
            val, weights=pw[j[0]], bins=100, histtype="step", label=label[j[0]]
        )
        ax.hist(
            val, weights=pw[j[1]], bins=100, histtype="step", label=label[j[1]]
        )
        x, y, _ = ax.hist(
            val,
            weights=pw_if[j],
            bins=100,
            histtype="step",
            label=label[j[0]] + "+" + label[j[1]],
        )
        ax.hist(
            val,
            weights=pw_if[j] - pw[j[0]] - pw[j[1]],
            bins=100,
            histtype="step",
            label=label[j[0]] + "x" + label[j[1]],
        )
        ax.plot(y, np.zeros_like(y))
        ax.legend(loc="upper right")
        # ax.set_ylim((None, np.max(y) *1.2))
        ax.set_title("M(DsD*)")
        plt.savefig("fig/fig_{}_{}.pdf".format(*j))
    # plt.savefig("test_plot.png")
    # plt.show()


def main():
    import argparse

    parser = argparse.ArgumentParser(description="None")
    parser.add_argument("-i", default="init_params.json", dest="init_params")
    results = parser.parse_args()
    test_plot(results.init_params)


if __name__ == "__main__":
    main()



from warnings import warn

import numpy as np

# from numpy import xrange
# from scipy.optimize.linesearch import line_search_wolfe1, line_search_wolfe2
from scipy.optimize import OptimizeResult


class LineSearchWarning(RuntimeWarning):
    pass


message_dict = {
    0: "Optimization terminated successfully.",
    3: "Maximum number of function evaluations has been exceeded.",
    2: "Maximum number of iterations has been exceeded.",
    1: "Desired error not necessarily achieved due to precision loss.",
    4: "NaN result encountered.",
}


class Cached_FG:
    def __init__(self, f_g):
        self.f_g = f_g
        self.cached_fun = 0
        self.cached_grad = 0
        self.ncall = 0

    def __call__(self, x):
        f = self.fun(x)
        return f, self.cached_grad

    def fun(self, x):
        f, g = self.f_g(x)
        self.cached_x = x
        self.cached_grad = g
        self.cached_fun = f
        self.ncall += 1
        return f

    def grad(self, x):
        if not np.all(x == self.cached_x):
            self.fun(x)
        return self.cached_grad


class Seq:
    def __init__(self, size=5):
        self._cached = []
        self.size = size

    def get_max(self):
        return max(self._cached)

    def arg_max(self):
        max_item = self._cached[0]
        max_idx = 0
        for i, v in enumerate(self._cached):
            if max_item < v:
                max_idx = i
                max_item = v
        return max_idx

    def add(self, x):
        self._cached.append(x)
        if len(self._cached) > self.size:
            self._cached.pop(0)


def fmin_bfgs_f(
    f_g,
    x0,
    B0=None,
    M=2,
    gtol=1e-5,
    Delta=10.0,
    maxiter=None,
    callback=None,
    norm_ord=np.Inf,
    **_kwargs
):
    """test BFGS with nonmonote line search"""
    fk, gk = f_g(x0)
    if B0 is None:
        Bk = np.eye(len(x0))
    else:
        Bk = B0
    Hk = np.linalg.inv(Bk)
    maxiter = 200 * len(x0) if maxiter is None else maxiter
    xk = x0
    norm = lambda x: np.linalg.norm(x, ord=norm_ord)
    theta = 0.9
    C = 0.5
    k = 0
    old_old_fval = fk + np.linalg.norm(gk) / 2
    old_fval = fk
    f_s = Seq(M)
    f_s.add(fk)
    flag = 0
    re_search = 0
    for k in range(maxiter):
        if norm(gk) <= gtol:
            break
        dki = -np.dot(Hk, gk)
        try:
            pk = dki
            f = f_g.fun
            myfprime = f_g.grad
            gfk = gk
            old_fval = fk
            (
                alpha_k,
                fc,
                gc,
                old_fval,
                old_old_fval,
                gfkp1,
            ) = line_search_wolfe2(
                f, myfprime, xk, pk, gfk, f_s.get_max(), old_fval, old_old_fval
            )
        except Exception as e:
            print(e)
            re_search += 1
            xk = xk + dki
            fk, gk = f_g(xk)
            old_fval, old_old_fval = fk, old_fval
            f_s.add(fk)
            if re_search > 2:
                flag = 1
                break
            continue
        if alpha_k is None:
            print("alpha is None")
            xk = xk + dki
            fk, gk = f_g(xk)
            old_fval, old_old_fval = fk, old_fval
            f_s.add(fk)
            re_search += 1
            if re_search > 2:
                flag = 1
                break
            continue
        dki = alpha_k * pk
        # fki, gki = f_g(xk + dki)
        fki, gki = old_fval, gfkp1
        Aredk = fk - fki
        Predk = -(np.dot(gk, dki) + 0.5 * np.dot(np.dot(Bk, dki), dki))
        rk = Aredk / Predk
        xk = xk + dki
        fk = fki
        yk = gki - gk
        tk = C + max(0, -np.dot(yk, dki) / norm(dki) ** 2) / norm(gk)
        ystark = (1 - theta) * yk + theta * tk * norm(gk) * dki
        gk = gki
        bs = np.dot(Bk, dki)
        Bk = (
            Bk
            + np.outer(yk, yk) / np.dot(yk, dki)
            - np.outer(bs, bs) / np.dot(bs, dki)
        )
        # sk = dki
        # rhok = 1.0 / (np.dot(yk, sk))
        # A1 = 1 - np.outer(sk, yk) * rhok
        # A2 = 1 - np.outer(yk, sk) * rhok
        # Hk = np.dot(A2, np.dot(Hk, A1)) - (rhok * np.outer(sk, sk))
        # Bk = Bk + np.outer(ystark, ystark)/np.dot(ystark, dki) - \
        #    np.outer(bs, bs)/np.dot(bs, dki)  # MBFGS
        # print(np.dot(Hk, Bk))
        try:
            Hk = np.linalg.inv(Bk)
        except Exception:
            pass
        f_s.add(fk)
        if callback is not None:
            callback(xk)
    else:
        flag = 2
    # print("fit final: ", k, p, f_g.ncall)
    s = OptimizeResult()
    s.messgae = message_dict[flag]
    s.fun = float(fk)
    s.nit = k
    s.nfev = f_g.ncall
    s.njev = f_g.ncall
    s.status = flag
    s.x = np.array(xk)
    s.jac = np.array(gk)
    s.hess = np.array(Bk)
    s.success = flag == 0
    return s


def minimize(
    fun,
    x0,
    args=(),
    method=None,
    jac=None,
    hess=None,
    hessp=None,
    bounds=None,
    constraints=(),
    tol=None,
    callback=None,
    options=None,
):
    options = options if options is not None else {}
    s = fmin_bfgs_f(Cached_FG(fun), x0, callback=callback, **options)
    return s


def line_search_nonmonote(
    f,
    myfprime,
    xk,
    pk,
    gfk=None,
    old_fval=None,
    fk=None,
    old_old_fval=None,
    args=(),
    c1=0.5,
    maxiter=10,
):
    alpha = max(-np.dot(gfk, pk) / np.dot(pk, pk), 1.0)
    phi_star = None
    print("init alpha", alpha, "\ngrad:", gfk)
    for i in range(maxiter):
        phi_star = f(xk + alpha * pk)
        if phi_star < old_fval + c1 * alpha * np.dot(gfk, pk):
            derphi_star = myfprime(xk + alpha * pk)
            return alpha, 0, 0, phi_star, old_fval, derphi_star
        alpha = c1 * alpha
    derphi_star = myfprime(xk + alpha * pk)
    print("not found")
    return alpha, 0, 0, phi_star, old_fval, derphi_star


# ------------------------------------------------------------------------------
# Pure-Python Wolfe line and scalar searches
# from scipy.optimize.linesearch
# ------------------------------------------------------------------------------


def line_search_wolfe2(
    f,
    myfprime,
    xk,
    pk,
    gfk=None,
    fk=None,
    old_fval=None,
    old_old_fval=None,
    args=(),
    c1=1e-4,
    c2=0.9,
    amax=None,
    extra_condition=None,
    maxiter=10,
):
    """Find alpha that satisfies strong Wolfe conditions.

    Parameters
    ----------
    f : callable f(x,*args)
        Objective function.
    myfprime : callable f'(x,*args)
        Objective function gradient.
    xk : ndarray
        Starting point.
    pk : ndarray
        Search direction.
    gfk : ndarray, optional
        Gradient value for x=xk (xk being the current parameter
        estimate). Will be recomputed if omitted.
    old_fval : float, optional
        Function value for x=xk. Will be recomputed if omitted.
    old_old_fval : float, optional
        Function value for the point preceding x=xk
    args : tuple, optional
        Additional arguments passed to objective function.
    c1 : float, optional
        Parameter for Armijo condition rule.
    c2 : float, optional
        Parameter for curvature condition rule.
    amax : float, optional
        Maximum step size
    extra_condition : callable, optional
        A callable of the form ``extra_condition(alpha, x, f, g)``
        returning a boolean. Arguments are the proposed step ``alpha``
        and the corresponding ``x``, ``f`` and ``g`` values. The line search
        accepts the value of ``alpha`` only if this
        callable returns ``True``. If the callable returns ``False``
        for the step length, the algorithm will continue with
        new iterates. The callable is only called for iterates
        satisfying the strong Wolfe conditions.
    maxiter : int, optional
        Maximum number of iterations to perform

    Returns
    -------
    alpha : float or None
        Alpha for which ``x_new = x0 + alpha * pk``,
        or None if the line search algorithm did not converge.
    fc : int
        Number of function evaluations made.
    gc : int
        Number of gradient evaluations made.
    new_fval : float or None
        New function value ``f(x_new)=f(x0+alpha*pk)``,
        or None if the line search algorithm did not converge.
    old_fval : float
        Old function value ``f(x0)``.
    new_slope : float or None
        The local slope along the search direction at the
        new value ``<myfprime(x_new), pk>``,
        or None if the line search algorithm did not converge.


    Notes
    -----
    Uses the line search algorithm to enforce strong Wolfe
    conditions.  See Wright and Nocedal, 'Numerical Optimization',
    1999, pg. 59-60.

    For the zoom phase it uses an algorithm by [...].

    """
    fc = [0]
    gc = [0]
    gval = [None]
    gval_alpha = [None]

    def phi(alpha):
        fc[0] += 1
        return f(xk + alpha * pk, *args)

    if isinstance(myfprime, tuple):

        def derphi(alpha):
            fc[0] += len(xk) + 1
            eps = myfprime[1]
            fprime = myfprime[0]
            newargs = (f, eps) + args
            gval[0] = fprime(xk + alpha * pk, *newargs)  # store for later use
            gval_alpha[0] = alpha
            return np.dot(gval[0], pk)

    else:
        fprime = myfprime

        def derphi(alpha):
            gc[0] += 1
            gval[0] = fprime(xk + alpha * pk, *args)  # store for later use
            gval_alpha[0] = alpha
            return np.dot(gval[0], pk)

    if gfk is None:
        gfk = fprime(xk, *args)
    derphi0 = np.dot(gfk, pk)

    if extra_condition is not None:
        # Add the current gradient as argument, to avoid needless
        # re-evaluation
        def extra_condition2(alpha, phi):
            if gval_alpha[0] != alpha:
                derphi(alpha)
            x = xk + alpha * pk
            return extra_condition(alpha, x, phi, gval[0])

    else:
        extra_condition2 = None

    alpha_star, phi_star, old_fval, derphi_star = scalar_search_wolfe2(
        phi,
        derphi,
        old_fval,
        old_old_fval,
        derphi0,
        c1,
        c2,
        amax,
        extra_condition2,
        maxiter=maxiter,
    )

    if derphi_star is None:
        warn("The line search algorithm did not converge", LineSearchWarning)
        return line_search_nonmonote(
            f,
            myfprime,
            xk,
            pk,
            gfk,
            fk,
            old_fval,
            old_old_fval,
            args,
            c1,
            maxiter,
        )
    else:
        # derphi_star is a number (derphi) -- so use the most recently
        # calculated gradient used in computing it derphi = gfk*pk
        # this is the gradient at the next step no need to compute it
        # again in the outer loop.
        derphi_star = gval[0]

    return alpha_star, fc[0], gc[0], phi_star, old_fval, derphi_star


def scalar_search_wolfe2(
    phi,
    derphi,
    phi0=None,
    old_phi0=None,
    derphi0=None,
    c1=1e-4,
    c2=0.9,
    amax=None,
    extra_condition=None,
    maxiter=10,
):
    """Find alpha that satisfies strong Wolfe conditions.

    alpha > 0 is assumed to be a descent direction.

    Parameters
    ----------
    phi : callable phi(alpha)
        Objective scalar function.
    derphi : callable phi'(alpha)
        Objective function derivative. Returns a scalar.
    phi0 : float, optional
        Value of phi at 0
    old_phi0 : float, optional
        Value of phi at previous point
    derphi0 : float, optional
        Value of derphi at 0
    c1 : float, optional
        Parameter for Armijo condition rule.
    c2 : float, optional
        Parameter for curvature condition rule.
    amax : float, optional
        Maximum step size
    extra_condition : callable, optional
        A callable of the form ``extra_condition(alpha, phi_value)``
        returning a boolean. The line search accepts the value
        of ``alpha`` only if this callable returns ``True``.
        If the callable returns ``False`` for the step length,
        the algorithm will continue with new iterates.
        The callable is only called for iterates satisfying
        the strong Wolfe conditions.
    maxiter : int, optional
        Maximum number of iterations to perform

    Returns
    -------
    alpha_star : float or None
        Best alpha, or None if the line search algorithm did not converge.
    phi_star : float
        phi at alpha_star
    phi0 : float
        phi at 0
    derphi_star : float or None
        derphi at alpha_star, or None if the line search algorithm
        did not converge.

    Notes
    -----
    Uses the line search algorithm to enforce strong Wolfe
    conditions.  See Wright and Nocedal, 'Numerical Optimization',
    1999, pg. 59-60.

    For the zoom phase it uses an algorithm by [...].

    """

    if phi0 is None:
        phi0 = phi(0.0)

    if derphi0 is None:
        derphi0 = derphi(0.0)

    alpha0 = 0
    if old_phi0 is not None and derphi0 != 0:
        alpha1 = min(1.0, 1.01 * 2 * (phi0 - old_phi0) / derphi0)
    else:
        alpha1 = 1.0

    if alpha1 < 0:
        alpha1 = 1.0

    if amax is not None:
        alpha1 = min(alpha1, amax)

    phi_a1 = phi(alpha1)
    # derphi_a1 = derphi(alpha1)  evaluated below

    phi_a0 = phi0
    derphi_a0 = derphi0

    if extra_condition is None:
        extra_condition = lambda alpha, phi: True

    for i in range(maxiter):
        if alpha1 == 0 or (amax is not None and alpha0 == amax):
            # alpha1 == 0: This shouldn't happen. Perhaps the increment has
            # slipped below machine precision?
            alpha_star = None
            phi_star = phi0
            phi0 = old_phi0
            derphi_star = None

            if alpha1 == 0:
                msg = "Rounding errors prevent the line search from converging"
            else:
                msg = (
                    "The line search algorithm could not find a solution "
                    + "less than or equal to amax: %s" % amax
                )

            warn(msg, LineSearchWarning)
            break

        if (phi_a1 > phi0 + c1 * alpha1 * derphi0) or (
            (phi_a1 >= phi_a0) and (i > 1)
        ):
            alpha_star, phi_star, derphi_star = _zoom(
                alpha0,
                alpha1,
                phi_a0,
                phi_a1,
                derphi_a0,
                phi,
                derphi,
                phi0,
                derphi0,
                c1,
                c2,
                extra_condition,
            )
            break

        derphi_a1 = derphi(alpha1)
        if abs(derphi_a1) <= -c2 * derphi0:
            if extra_condition(alpha1, phi_a1):
                alpha_star = alpha1
                phi_star = phi_a1
                derphi_star = derphi_a1
                break

        if derphi_a1 >= 0:
            alpha_star, phi_star, derphi_star = _zoom(
                alpha1,
                alpha0,
                phi_a1,
                phi_a0,
                derphi_a1,
                phi,
                derphi,
                phi0,
                derphi0,
                c1,
                c2,
                extra_condition,
            )
            break

        alpha2 = 2 * alpha1  # increase by factor of two on each iteration
        if amax is not None:
            alpha2 = min(alpha2, amax)
        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi(alpha1)
        derphi_a0 = derphi_a1

    else:
        # stopping test maxiter reached
        alpha_star = alpha1
        phi_star = phi_a1
        derphi_star = None
        warn("The line search algorithm did not converge", LineSearchWarning)

    return alpha_star, phi_star, phi0, derphi_star


def _cubicmin(a, fa, fpa, b, fb, c, fc):
    """
    Finds the minimizer for a cubic polynomial that goes through the
    points (a,fa), (b,fb), and (c,fc) with derivative at a of fpa.

    If no minimizer can be found return None

    """
    # f(x) = A *(x-a)^3 + B*(x-a)^2 + C*(x-a) + D

    with np.errstate(divide="raise", over="raise", invalid="raise"):
        try:
            C = fpa
            db = b - a
            dc = c - a
            denom = (db * dc) ** 2 * (db - dc)
            d1 = np.empty((2, 2))
            d1[0, 0] = dc ** 2
            d1[0, 1] = -(db ** 2)
            d1[1, 0] = -(dc ** 3)
            d1[1, 1] = db ** 3
            [A, B] = np.dot(
                d1, np.asarray([fb - fa - C * db, fc - fa - C * dc]).flatten()
            )
            A /= denom
            B /= denom
            radical = B * B - 3 * A * C
            xmin = a + (-B + np.sqrt(radical)) / (3 * A)
        except ArithmeticError:
            return None
    if not np.isfinite(xmin):
        return None
    return xmin


def _quadmin(a, fa, fpa, b, fb):
    """
    Finds the minimizer for a quadratic polynomial that goes through
    the points (a,fa), (b,fb) with derivative at a of fpa,

    """
    # f(x) = B*(x-a)^2 + C*(x-a) + D
    with np.errstate(divide="raise", over="raise", invalid="raise"):
        try:
            D = fa
            C = fpa
            db = b - a * 1.0
            B = (fb - D - C * db) / (db * db)
            xmin = a - C / (2.0 * B)
        except ArithmeticError:
            return None
    if not np.isfinite(xmin):
        return None
    return xmin


def _zoom(
    a_lo,
    a_hi,
    phi_lo,
    phi_hi,
    derphi_lo,
    phi,
    derphi,
    phi0,
    derphi0,
    c1,
    c2,
    extra_condition,
):
    """
    Part of the optimization algorithm in `scalar_search_wolfe2`.
    """

    maxiter = 10
    i = 0
    delta1 = 0.2  # cubic interpolant check
    delta2 = 0.1  # quadratic interpolant check
    phi_rec = phi0
    a_rec = 0
    while True:
        # interpolate to find a trial step length between a_lo and
        # a_hi Need to choose interpolation here.  Use cubic
        # interpolation and then if the result is within delta *
        # dalpha or outside of the interval bounded by a_lo or a_hi
        # then use quadratic interpolation, if the result is still too
        # close, then use bisection

        dalpha = a_hi - a_lo
        if dalpha < 0:
            a, b = a_hi, a_lo
        else:
            a, b = a_lo, a_hi

        # minimizer of cubic interpolant
        # (uses phi_lo, derphi_lo, phi_hi, and the most recent value of phi)
        #
        # if the result is too close to the end points (or out of the
        # interval) then use quadratic interpolation with phi_lo,
        # derphi_lo and phi_hi if the result is still too close to the
        # end points (or out of the interval) then use bisection

        if i > 0:
            cchk = delta1 * dalpha
            a_j = _cubicmin(
                a_lo, phi_lo, derphi_lo, a_hi, phi_hi, a_rec, phi_rec
            )
        if (i == 0) or (a_j is None) or (a_j > b - cchk) or (a_j < a + cchk):
            qchk = delta2 * dalpha
            a_j = _quadmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi)
            if (a_j is None) or (a_j > b - qchk) or (a_j < a + qchk):
                a_j = a_lo + 0.5 * dalpha

        # Check new value of a_j

        phi_aj = phi(a_j)
        if (phi_aj > phi0 + c1 * a_j * derphi0) or (phi_aj >= phi_lo):
            phi_rec = phi_hi
            a_rec = a_hi
            a_hi = a_j
            phi_hi = phi_aj
        else:
            derphi_aj = derphi(a_j)
            if abs(derphi_aj) <= -c2 * derphi0 and extra_condition(
                a_j, phi_aj
            ):
                a_star = a_j
                val_star = phi_aj
                valprime_star = derphi_aj
                break
            if derphi_aj * (a_hi - a_lo) >= 0:
                phi_rec = phi_hi
                a_rec = a_hi
                a_hi = a_lo
                phi_hi = phi_lo
            else:
                phi_rec = phi_lo
                a_rec = a_lo
            a_lo = a_j
            phi_lo = phi_aj
            derphi_lo = derphi_aj
        i += 1
        if i > maxiter:
            # Failed to find a conforming step size
            a_star = None
            val_star = None
            valprime_star = None
            break
    return a_star, val_star, valprime_star



#!/usr/bin/env python3
import copy
import os.path
import sys
from pprint import pprint

import yaml

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.significance import significance

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def single_fit(config_dict, data, phsp, bg):
    config = ConfigLoader(config_dict)

    print("\n########### initial parameters")
    pprint(config.get_params())
    print(config.full_decay)
    fit_result = config.fit(data, phsp, bg=bg)
    pprint(fit_result.params)
    # fit_result.save_as("final_params.json")
    return fit_result.min_nll, fit_result.ndf


def multi_fit(config, data, phsp, bg, num=5):
    nll, ndf = single_fit(config, data, phsp, bg)
    for i in range(num - 1):
        nll_i, ndf_i = single_fit(config, data, phsp, bg)
        assert ndf_i == ndf
        if nll > nll_i:
            nll = nll_i
    return nll, ndf


def veto_resonance(config, res):
    config = copy.deepcopy(config)
    for k, v in config["particle"].items():
        if isinstance(v, list):
            if res in v:
                config["particle"][k].remove(res)
    if res in config["decay"]:
        config["decay"].remove(res)
    for k, v in config["decay"].items():
        # if res in v:
        #     decay[k].remove(res)
        for j in v:
            if isinstance(j, list):
                if res in j:
                    j.remove(res)
    return config


def cached_data(config_dict):
    config = ConfigLoader(config_dict)
    data, phsp, bg = config.get_all_data()[:3]
    return data, phsp, bg


def cal_significance(config_name, res, model="-"):
    with open(config_name) as f:
        config = yaml.safe_load(f)
    data, phsp, bg = cached_data(config)

    def get_config(extra=[]):
        base_conf = copy.deepcopy(config)
        if model == "+":
            veto_res = res.copy()
            for i in extra:
                if i in veto_res:
                    veto_res.remove(i)
        else:
            veto_res = extra
        for i in veto_res:
            base_conf = veto_resonance(base_conf, i)
        return base_conf

    base_config = get_config([])
    nll, ndf = multi_fit(base_config, data, phsp, bg)
    nlls = {"base": nll}
    ndfs = {"base": ndf}
    print("nll: {}, ndf: {}".format(nll, ndf))
    signi = {}
    for i in res:
        print("\ncalculate significance for {}\n".format(i))
        config_i = get_config([i])
        nll_i, ndf_i = multi_fit(config_i, data, phsp, bg)
        nlls[i] = nll_i
        ndfs[i] = ndf_i
        signi[i] = significance(nll, nll_i, abs(ndf - ndf_i))
        print(
            "nll: {}, ndf: {}, significane: {}".format(nll_i, ndf_i, signi[i])
        )
    return signi, nlls, ndfs


def main():
    res = [
        "X(3940)(1+)",
        "X(3940)(1-)",
        "X(3940)(0-)",
        "X(3940)(2+)",
        "X(3940)(2-)",
        "Psi(4660)",
        "Psi(4230)",
        "Psi(4390)",
        "Psi(4260)",
        "Psi(4360)",
    ]
    import argparse

    parser = argparse.ArgumentParser(description="calculate significance")
    parser.add_argument("--config", default="config.yml", dest="config")
    results = parser.parse_args()
    signi, nlls, ndfs = cal_significance(results.config, res, model="+")
    print("base", nlls["base"], ndfs["base"])
    print("particle\tsignificance\tnll\tndf")
    for i in signi:
        print(i, signi[i], nlls[i], ndfs[i])


if __name__ == "__main__":
    main()



import os
import sys

import pytest

from .test_full import gen_toy


def test_fit(gen_toy):
    print(os.getcwd())
    sys.argv = [
        "fit.py",
        "-c",
        "tf_pwa/tests/config_toy.yml",
        "-i",
        "tf_pwa/tests/exp_params.json",
        "--no-GPU",
    ]
    exec("from fit import main; main()")


def test_main_fit(gen_toy):
    from tf_pwa.app.fit import fit

    fit("tf_pwa/tests/config_toy.yml", "tf_pwa/tests/exp_params.json")



import yaml

from tf_pwa.config_loader import ConfigLoader

from .test_full import gen_toy, this_dir


def test_cp_decay():
    with open(f"{this_dir}/config_toy.yml") as f:
        config_data = yaml.full_load(f)
    config_data["decay_chain"] = {"$all": {"is_cp": True}}
    config = ConfigLoader(config_data)

    amp = config.get_amplitude()
    data = config.get_data("data")[0]
    amp(data)



import numpy as np
import tensorflow as tf

from tf_pwa.data import data_shape
from tf_pwa.experimental import build_amp, opt_int

from .model import (
    Model,
    _loop_generator,
    clip_log,
    split_generator,
    sum_hessian,
)


def sum_gradient(fs, var, weight=1.0, trans=tf.identity, args=(), kwargs=None):
    """
    NLL is the sum of trans(f(data)):math:`*`weight; gradient is the derivatives for each variable in ``var``.

    :param f: Function. The amplitude PDF.
    :param var: List of strings. Names of the trainable variables in the PDF.
    :param weight: Weight factor for each data point. It's either a real number or an array of the same shape with ``data``.
    :param trans: Function. Transformation of ``data`` before multiplied by ``weight``.
    :param kwargs: Further arguments for ``f``.
    :return: Real number NLL, list gradient
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    ys = []
    gs = []
    for f, weight_i in zip(fs, weight):
        with tf.GradientTape() as tape:
            part_y = trans(f(*args, **kwargs))
            y_i = tf.reduce_sum(tf.cast(weight_i, part_y.dtype) * part_y)
        g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
        ys.append(y_i)
        gs.append(g_i)
    nll = sum(ys)
    g = list(map(sum, zip(*gs)))
    return nll, g


def sum_gradient_data2(
    f,
    var,
    data,
    cached_data,
    weight=1.0,
    trans=tf.identity,
    args=(),
    kwargs=None,
):
    """
    NLL is the sum of trans(f(data)):math:`*`weight; gradient is the derivatives for each variable in ``var``.

    :param f: Function. The amplitude PDF.
    :param var: List of strings. Names of the trainable variables in the PDF.
    :param weight: Weight factor for each data point. It's either a real number or an array of the same shape with ``data``.
    :param trans: Function. Transformation of ``data`` before multiplied by ``weight``.
    :param kwargs: Further arguments for ``f``.
    :return: Real number NLL, list gradient
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    ys = []
    gs = []
    for data_i, c_data_i, weight_i in zip(data, cached_data, weight):
        with tf.GradientTape() as tape:
            part_y = trans(f(data_i, c_data_i, *args, **kwargs))
            y_i = tf.reduce_sum(tf.cast(weight_i, part_y.dtype) * part_y)
        g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
        ys.append(y_i)
        gs.append(g_i)
    nll = sum(ys)
    g = list(map(sum, zip(*gs)))
    return nll, g


def sum_grad_hessp_data2(
    f,
    p,
    var,
    data,
    data2,
    weight=1.0,
    trans=tf.identity,
    resolution_size=1,
    args=(),
    kwargs=None,
):
    """
    The parameters are the same with ``sum_gradient()``, but this function will return hessian as well,
    which is the matrix of the second-order derivative.

    :return: Real number NLL, list gradient, 2-D list hessian
    """
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(weight, float):
        weight = _loop_generator(weight)
    y_s = []
    g_s = []
    h_s = []
    from tensorflow.python.eager import forwardprop

    for data_i, c_data_i, weight_i in zip(data, data2, weight):
        with forwardprop.ForwardAccumulator(var, list(p)) as acc:
            with tf.GradientTape() as tape:
                part_y = trans(f(data_i, c_data_i, *args, **kwargs))
                y_i = tf.reduce_sum(tf.cast(weight_i, part_y.dtype) * part_y)
            g_i = tape.gradient(y_i, var, unconnected_gradients="zero")
        hessp = acc.jvp(g_i, unconnected_gradients="zero")
        y_s.append(y_i)
        g_s.append(g_i)
        h_s.append(hessp)
        # print(hessp)
    nll = tf.reduce_sum(y_s)
    # print("ll: ", nll)
    g = tf.reduce_sum(g_s, axis=0)
    h = tf.reduce_sum(h_s, axis=0)
    # print(h)
    # h = [[sum(j) for j in zip(*i)] for i in h_s]
    return nll, g, h


class ModelCachedInt(Model):
    """
    This class implements methods to calculate NLL as well as its derivatives for an amplitude model with Cached Int.
    It may include data for both signal and background.
    Cached Int well cause wrong results when float parameters include mass or width.

    :param amp: ``AllAmplitude`` object. The amplitude model.
    :param w_bkg: Real number. The weight of background.
    """

    def __init__(self, amp, w_bkg=1.0):
        super(ModelCachedInt, self).__init__(amp, w_bkg)
        self.cached_int = {}
        self.cached_amp = {}

    def build_cached_int(self, mcdata, mc_weight, batch=65000):
        mc_id = id(mcdata)
        if isinstance(mcdata, dict):
            mcdata = split_generator(mcdata, batch)
            mc_weight = split_generator(mc_weight, batch)
        dec = self.Amp.decay_group
        index, ret = None, []
        sum_weight = 1.0
        for data, weight in zip(mcdata, mc_weight):
            sum_weight += tf.reduce_sum(weight)
            index, a = opt_int.build_int_matrix(dec, data, weight)
            ret.append(a)

        int_matrix = tf.reduce_sum(ret, axis=0)

        @tf.function
        def int_mc():
            pm = opt_int.build_params_matrix(dec)
            ret = tf.reduce_sum(pm * int_matrix)
            return tf.math.real(ret)

        self.cached_int[mc_id] = int_mc

        # print(int_mc())
        # a = 0.0
        # for mc, w in zip(mcdata, mc_weight):
        #     a += tf.reduce_sum(self.Amp(mc) * w)
        # print(a)

    def get_cached_int(self, mc_id):
        return self.cached_int[mc_id]()

    # @tf.function
    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
        data_id = id(data)
        data = list(data)
        weight = list(weight)
        if data_id not in self.cached_amp:
            self.cached_amp[data_id] = [
                build_amp.cached_amp2s(self.Amp.decay_group, i) for i in data
            ]
        ln_data, g_ln_data = sum_gradient(
            self.cached_amp[data_id],
            self.Amp.trainable_variables,
            weight=weight,
            trans=clip_log,
        )
        # print(ln_data, ln_data2, np.allclose(g_ln_data, g_ln_data2))
        mc_id = id(mcdata)
        if mc_id not in self.cached_int:
            self.build_cached_int(mcdata, mc_weight)
        with tf.GradientTape() as tape:
            int_mc = self.get_cached_int(mc_id)
        g_int_mc = tape.gradient(
            int_mc, self.Amp.trainable_variables, unconnected_gradients="zero"
        )

        # int_mc2, g_int_mc2 = sum_gradient(self.Amp, mcdata,
        #                                self.Amp.trainable_variables, weight=mc_weight)
        #
        # print("exp", int_mc, g_int_mc)
        # print("now", int_mc2, g_int_mc2)
        sw = tf.cast(sw, ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )
        nll = -ln_data + sw * tf.math.log(int_mc)

        return nll, g

    def nll_grad_hessian(
        self, data, mcdata, weight=1.0, batch=24000, bg=None, mc_weight=1.0
    ):
        """
        The parameters are the same with ``self.nll()``, but it will return Hessian as well.

        :return NLL: Real number. The value of NLL.
        :return gradients: List of real numbers. The gradients for each variable.
        :return Hessian: 2-D Array of real numbers. The Hessian matrix of the variables.
        """
        data, weight = self.get_weight_data(data, weight, bg=bg)
        if isinstance(mc_weight, float):
            mc_weight = tf.convert_to_tensor(
                [mc_weight] * data_shape(mcdata), dtype="float64"
            )
        n_mc = tf.reduce_sum(mc_weight)
        sw = tf.reduce_sum(weight)
        ln_data, g_ln_data, h_ln_data = sum_hessian(
            self.Amp,
            split_generator(data, batch),
            self.Amp.trainable_variables,
            weight=split_generator(weight, batch),
            trans=clip_log,
        )

        # int_mc, g_int_mc, h_int_mc = sum_hessian(self.Amp, split_generator(mcdata, batch),
        #                                         self.Amp.trainable_variables, weight=split_generator(
        #        mc_weight, batch))
        if isinstance(mc_weight, float):
            mc_weight = tf.convert_to_tensor(
                [mc_weight] * data_shape(mcdata), dtype="float64"
            )
            mc_weight = mc_weight / tf.reduce_sum(mc_weight)
        mc_id = id(mcdata)
        if mc_id not in self.cached_int:
            self.build_cached_int(mcdata, mc_weight)
        with tf.GradientTape(persistent=True) as tape0:
            with tf.GradientTape() as tape:
                y_i = self.get_cached_int(mc_id)
            g_i = tape.gradient(
                y_i, self.Amp.trainable_variables, unconnected_gradients="zero"
            )
        h_s_i = []
        for gi in g_i:
            # 2nd order derivative
            h_s_i.append(
                tape0.gradient(
                    gi,
                    self.Amp.trainable_variables,
                    unconnected_gradients="zero",
                )
            )
        del tape0
        int_mc = y_i
        g_int_mc = tf.convert_to_tensor(g_i)
        h_int_mc = tf.convert_to_tensor(h_s_i)

        n_var = len(g_ln_data)
        nll = -ln_data + sw * tf.math.log(int_mc / n_mc)
        g = -g_ln_data + sw * g_int_mc / int_mc

        g_int_mc = g_int_mc / int_mc
        g_outer = tf.reshape(g_int_mc, (-1, 1)) * tf.reshape(g_int_mc, (1, -1))

        h = -h_ln_data - sw * g_outer + sw / int_mc * h_int_mc
        return nll, g, h


class ModelCachedAmp(Model):
    """
    This class implements methods to calculate NLL as well as its derivatives for an amplitude model with Cached Int.
    It may include data for both signal and background.
    Cached Int well cause wrong results when float parameters include mass or width.

    :param amp: ``AllAmplitude`` object. The amplitude model.
    :param w_bkg: Real number. The weight of background.
    """

    def __init__(self, amp, w_bkg=1.0):
        super(ModelCachedAmp, self).__init__(amp, w_bkg)
        self.cached_amp = build_amp.build_amp2s(amp.decay_group)
        self.cached_data = {}

    def sum_nll_grad_bacth(self, data):
        data_id = id(data)
        data = list(data)
        weight = [i.get("weight", tf.ones((data_shape(i),))) for i in data]
        if data_id not in self.cached_data:
            self.cached_data[data_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in data
            ]
        ln_data, g_ln_data = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            data,
            self.cached_data[data_id],
            weight=weight,
            trans=clip_log,
        )
        return -ln_data, [-i for i in g_ln_data]

    def sum_log_integral_grad_batch(self, mcdata, ndata):
        mc_id = id(mcdata)
        mcdata = list(mcdata)
        mc_weight = [i["weight"] for i in mcdata]
        if mc_id not in self.cached_data:
            self.cached_data[mc_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in mcdata
            ]
        int_mc, g_int_mc = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            mcdata,
            self.cached_data[mc_id],
            weight=mc_weight,
        )
        return tf.math.log(int_mc) * ndata, [
            ndata / int_mc * i for i in g_int_mc
        ]

        # @tf.function

    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
        data_id = id(data)
        data = list(data)
        weight = list(weight)
        if data_id not in self.cached_data:
            self.cached_data[data_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in data
            ]
        ln_data, g_ln_data = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            data,
            self.cached_data[data_id],
            weight=weight,
            trans=clip_log,
        )
        # print(ln_data, ln_data2, np.allclose(g_ln_data, g_ln_data2))
        mc_id = id(mcdata)
        mcdata = list(mcdata)
        if mc_id not in self.cached_data:
            self.cached_data[mc_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in mcdata
            ]
        int_mc, g_int_mc = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            mcdata,
            self.cached_data[mc_id],
            weight=mc_weight,
        )

        # int_mc2, g_int_mc2 = sum_gradient(self.Amp, mcdata,
        #                                self.Amp.trainable_variables, weight=mc_weight)
        #
        # print("exp", int_mc, g_int_mc)
        # print("now", int_mc2, g_int_mc2)
        sw = tf.cast(sw, ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )
        nll = -ln_data + sw * tf.math.log(int_mc)

        return nll, g

    # @tf.function
    def nll_grad_batch(self, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
        data_id = id(data)
        data = list(data)
        weight = list(weight)
        if data_id not in self.cached_data:
            self.cached_data[data_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in data
            ]
        ln_data, g_ln_data = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            data,
            self.cached_data[data_id],
            weight=weight,
            trans=clip_log,
        )
        # print(ln_data, ln_data2, np.allclose(g_ln_data, g_ln_data2))
        mc_id = id(mcdata)
        mcdata = list(mcdata)
        if mc_id not in self.cached_data:
            self.cached_data[mc_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in mcdata
            ]
        int_mc, g_int_mc = sum_gradient_data2(
            self.cached_amp,
            self.Amp.trainable_variables,
            mcdata,
            self.cached_data[mc_id],
            weight=mc_weight,
        )

        # int_mc2, g_int_mc2 = sum_gradient(self.Amp, mcdata,
        #                                self.Amp.trainable_variables, weight=mc_weight)
        #
        # print("exp", int_mc, g_int_mc)
        # print("now", int_mc2, g_int_mc2)
        sw = tf.cast(sw, ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )
        nll = -ln_data + sw * tf.math.log(int_mc)

        return nll, g

    def grad_hessp_batch(self, p, data, mcdata, weight, mc_weight):
        """
        ``self.nll_grad()`` is replaced by this one???

        .. math::
          - \\frac{\\partial \\ln L}{\\partial \\theta_k } =
            -\\sum_{x_i \\in data } w_i \\frac{\\partial}{\\partial \\theta_k} \\ln f(x_i;\\theta_k)
            + (\\sum w_j ) \\left( \\frac{ \\partial }{\\partial \\theta_k} \\sum_{x_i \\in mc} f(x_i;\\theta_k) \\right)
              \\frac{1}{ \\sum_{x_i \\in mc} f(x_i;\\theta_k) }

        :param data:
        :param mcdata:
        :param weight:
        :param mc_weight:
        :return:
        """
        if not hasattr(self, "hess_product_vector_i"):
            self.hess_product_vector_i = [tf.Variable(i) for i in p]
        for i, j in zip(self.hess_product_vector_i, p):
            i.assign(j)
        data_id = id(data)
        data = list(data)
        weight = list(weight)
        sw = tf.reduce_sum([tf.reduce_sum(i) for i in weight])
        if data_id not in self.cached_data:
            self.cached_data[data_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in data
            ]
        # print(ln_data, ln_data2, np.allclose(g_ln_data, g_ln_data2))
        mc_id = id(mcdata)
        mcdata = list(mcdata)
        if mc_id not in self.cached_data:
            self.cached_data[mc_id] = [
                build_amp.build_angle_amp_matrix(self.Amp.decay_group, i)[1]
                for i in mcdata
            ]

        ln_data, g_ln_data, hessp_ln_data = sum_grad_hessp_data2(
            self.cached_amp,
            self.hess_product_vector_i,
            self.Amp.trainable_variables,
            data,
            self.cached_data[data_id],
            weight=weight,
            trans=clip_log,
            resolution_size=self.resolution_size,
        )

        # print("hessp_ln_data",hessp_ln_data)

        int_mc, g_int_mc, hessp_int_mc = sum_grad_hessp_data2(
            self.cached_amp,
            self.hess_product_vector_i,
            self.Amp.trainable_variables,
            mcdata,
            self.cached_data[mc_id],
            weight=mc_weight,
        )

        # print("hessp_int_mc", hessp_int_mc)

        sw = tf.cast(sw, ln_data.dtype)

        g = list(
            map(lambda x: -x[0] + sw * x[1] / int_mc, zip(g_ln_data, g_int_mc))
        )

        g_int_mc = np.array(g_int_mc)
        hessp2 = sw * (
            hessp_int_mc / int_mc
            - g_int_mc * np.dot(p, g_int_mc) / int_mc ** 2
        )
        # print("hessp2", hessp2)
        # print("ret", g, hessp2 - hessp_ln_data)
        return g, hessp2 - hessp_ln_data



import numpy as np

from tf_pwa.model.model import sum_gradient, sum_hessian
from tf_pwa.tensorflow_wrapper import tf


def test_sum_hessian():
    a = tf.Variable(1.0)
    b = tf.Variable(2.0)

    def f(x):
        return a + b ** 2 + x

    nll, g, h = sum_hessian(f, [1.0, 2.0], [a, b])
    nll1, g1 = sum_gradient(f, [1.0, 2.0], [a, b])
    assert nll == nll1
    assert np.allclose(g, g1)
    assert nll == 13.0
    assert np.allclose(g, np.array([2.0, 8.0]))
    assert np.allclose(h, np.array([[0.0, 0.0], [0.0, 4.0]]))



import itertools

import tensorflow as tf

from tf_pwa.amp import get_decay, get_particle
from tf_pwa.data import data_split


def split_gls(dec_chain):
    gls = [i.get_ls_list() for i in dec_chain]
    ls_combination = list(itertools.product(*gls))
    for i in ls_combination:
        for gi, j in zip(i, dec_chain):
            j.set_ls([gi])
        yield i, dec_chain
    for j, g in zip(dec_chain, gls):
        j.set_ls(g)


def build_sum_amplitude(dg, dec_chain, data):
    cached = []
    for i, dc in split_gls(dec_chain):
        amp = dg.get_amp(data)
        int_mc = amp
        gls = dc.product_gls()
        cached.append(amp / gls)
    return cached


def build_int_matrix(dec, data, weight=None):
    hij = {}
    used_chains = dec.chains_idx
    for k, i in enumerate(dec):
        dec.set_used_chains([k])
        for j, amp in enumerate(build_sum_amplitude(dec, i, data)):
            hij[(i, j)] = amp
    dec.set_used_chains(used_chains)
    ret = []
    if weight is None:
        weight = data.get("weight", 1.0)
    index = list(hij.keys())
    weight = tf.cast(weight, hij[index[0]].dtype)
    n_lambda = len(hij[index[0]].shape) - 1
    weight = tf.reshape(weight, [-1] + [1] * n_lambda)
    for i in index:
        tmp = []
        for j in index:
            xij = hij[i] * tf.math.conj(hij[j])
            xij = tf.reduce_sum(weight * xij)
            tmp.append(xij)
        ret.append(tmp)
    return index, ret


def build_int_matrix_batch(dec, data, batch=65000):
    index, ret = None, []
    for i in data_split(data, batch):
        index, a = build_int_matrix(dec, i)
        ret.append(a)
    return index, tf.reduce_sum(ret, axis=0)


def build_params_vector(dec):
    ret = []
    for i in dec:
        factor = i.get_all_factor()
        a = gls_combine(factor)
        ret.append(a)
    return tf.concat(ret, axis=0)


def build_params_matrix(dec):
    pv = build_params_vector(dec)
    return pv[:, None] * tf.math.conj(pv)[None, :]


def gls_combine(fs):
    ret = fs[0]
    for i in fs[1:]:
        ret = tf.reshape(ret[:, None] * i[None, :], (-1,))
    return ret


def cached_int_mc(dec, data, batch=65000):
    a, int_matrix = build_int_matrix_batch(dec, data, batch)

    @tf.function
    def int_mc():
        pm = build_params_matrix(dec)
        ret = tf.reduce_sum(pm * int_matrix)
        return tf.math.real(ret)

    return int_mc



#!/usr/bin/env python3
import json
import math
import os.path
import sys
import time

import numpy as np
import tensorflow as tf
from scipy.optimize import BFGS, basinhopping, minimize

from tf_pwa.amplitude import AllAmplitude
from tf_pwa.angle import cal_ang_file
from tf_pwa.fitfractions import cal_fitfractions
from tf_pwa.model import FCN, Cache_Model, param_list
from tf_pwa.utils import flatten_np_data, load_config_file

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


def error_print(x, err):
    return ("{:.7f} +/- {:.7f}").format(x, err)


param_list = [
    "m_A",
    "m_B",
    "m_C",
    "m_D",
    "m_BC",
    "m_BD",
    "m_CD",
    "beta_BC",
    "beta_B_BC",
    "alpha_BC",
    "alpha_B_BC",
    "beta_BD",
    "beta_B_BD",
    "alpha_BD",
    "alpha_B_BD",
    "beta_CD",
    "beta_D_CD",
    "alpha_CD",
    "alpha_D_CD",
    "beta_BD_B",
    "beta_BC_B",
    "beta_BD_D",
    "beta_CD_D",
    "alpha_BD_B",
    "gamma_BD_B",
    "alpha_BC_B",
    "gamma_BC_B",
    "alpha_BD_D",
    "gamma_BD_D",
    "alpha_CD_D",
    "gamma_CD_D",
]


def pprint(x):
    s = json.dumps(x, indent=2)
    print(s)


def part_config(config, name=[]):
    if isinstance(name, str):
        print(name)
        return {name: config[name]}
    ret = {}
    for i in name:
        ret[i] = config[i]
    return ret


def err_trans(grad, Vij):
    return np.dot(grad.T, np.dot(Vij, grad))


def main():
    dtype = "float64"
    # set_gpu_mem_growth()
    tf.keras.backend.set_floatx(dtype)
    config_list = load_config_file("Resonances")
    fname = [
        ["./data/data4600_new.dat", "data/Dst0_data4600_new.dat"],
        ["./data/bg4600_new.dat", "data/Dst0_bg4600_new.dat"],
        ["./data/PHSP4600_new.dat", "data/Dst0_PHSP4600_new.dat"],
        ["./data/PHSP_NOEFF.dat", ""],
    ]
    tname = ["data", "bg", "PHSP", "MC"]
    data_np = {}
    for i in range(len(fname)):
        data_np[tname[i]] = cal_ang_file(fname[i][0], dtype)

    def load_data(name):
        dat = []
        tmp = flatten_np_data(data_np[name])
        for i in param_list:
            tmp_data = tf.Variable(tmp[i], name=i, dtype=dtype)
            dat.append(tmp_data)
        return dat

    data = load_data("data")
    bg = load_data("bg")
    mcdata = load_data("PHSP")
    # flat_mc_data = load_data("MC")
    a = Cache_Model(config_list, 0.768331, data, mcdata, bg=bg, batch=26000)
    n_data = data[0].shape[0]
    sw = a.sw.numpy()

    with open("final_params.json") as f:
        param = json.load(f)
        if "value" in param:
            a.set_params(param["value"])
        else:
            a.set_params(param)
    s = json.dumps(a.get_params(), indent=2)
    print("params:")
    print(s)
    # np.savetxt("filename.txt",a)
    try:
        Vij = np.load("error_matrix.npy")
    except:
        (
            nll,
            g,
            h,
        ) = a.cal_nll_hessian()  # data_w,mcdata,weight=weights,batch=50000)
        inv_he = np.linalg.inv(h.numpy())
        Vij = inv_he
    mcdata_cached = a.Amp.cache_data(*mcdata, batch=65000)
    # flat_mc_data_cached = a.Amp.cache_data(*flat_mc_data,batch=65000)
    allvar = [i.name for i in a.Amp.trainable_variables]
    print(dict(zip(allvar, np.sqrt(np.diag(Vij).tolist()))))

    fitFrac, g_fitFrac = cal_fitfractions(
        a.Amp, mcdata_cached, kwargs={"cached": True}
    )
    # fitFrac_f,g_fitFrac_f = cal_fitfractions(a.Amp,flat_mc_data_cached,kwargs={"cached":True})
    print("check sum:", np.sum([fitFrac[i] for i in fitFrac]))
    print("fitfractions:")
    for i in fitFrac:
        s = error_print(fitFrac[i], np.sqrt(err_trans(g_fitFrac[i], Vij)))
        print(i, ":", s)

    # print("\nphysics fitfractions:")
    # for i in fitFrac_f:
    # s = error_print(fitFrac_f[i], np.sqrt(err_trans(g_fitFrac_f[i],Vij)))
    # print(i,":",s)


if __name__ == "__main__":
    main()



import functools

import tensorflow as tf


def extra_function(f0=None, using_numpy=True):
    """Using extra function with numerical differentiation.

    It can be used for numpy function or numba.vectorize function interface.

    >>> import numpy as np
    >>> sin2 = extra_function(np.sin)
    >>> a = tf.Variable([1.0,2.0], dtype="float64")
    >>> with tf.GradientTape(persistent=True) as tape0:
    ...     with tf.GradientTape(persistent=True) as tape:
    ...         b = sin2(a)
    ...     g, = tape.gradient(b, [a,])
    ...
    >>> h, = tape0.gradient(g, [a,])
    >>> assert np.allclose(np.sin([1.0,2.0]), b.numpy())
    >>> assert np.allclose(np.cos([1.0,2.0]), g.numpy())
    >>> assert np.sum(np.abs(-np.sin([1.0,2.0]) - h.numpy())) < 1e-3

    The numerical accuracy is not so well for second derivative.

    """

    def _wrapper(f):
        delta_x = {"float64": 1e-6, "float32": 1e-3}

        @tf.custom_gradient
        def _grad(x, **kwargs):
            if using_numpy and hasattr(x, "numpy"):
                x = x.numpy()
            h = delta_x[x.dtype.name]
            f_u = f(x + h, **kwargs)
            f_d = f(x - h, **kwargs)
            f_0 = f(x, **kwargs)

            def _hess(dg):
                tmp = (f_u + f_d - 2 * f_0) / h / h
                return dg * tmp

            return (f_u - f_d) / 2 / h, _hess

        @tf.custom_gradient
        @functools.wraps(f)
        def _f(x, **kwargs):
            def _g2(dy):
                return dy * _grad(x, **kwargs)

            if using_numpy and hasattr(x, "numpy"):
                x2 = x.numpy()
            else:
                x2 = x
            f_0 = f(x2)
            return f_0, _g2

        return _f

    if f0 is None:
        return _wrapper
    return _wrapper(f0)
